""" Extract parse tree(s) from a chart following a particular objective
function. """

from __future__ import print_function
import re
import logging
from heapq import nlargest
from math import exp, log, isinf, fsum
from random import random
from bisect import bisect_right
from operator import itemgetter, attrgetter
from itertools import count
from collections import defaultdict, OrderedDict
from discodop import plcfrs
from discodop.tree import Tree
from discodop.kbest import lazykbest, getderiv
from discodop.grammar import lcfrs_productions
from discodop.treetransforms import addbitsets, unbinarize, canonicalize
from discodop.agenda cimport Entry, new_Entry
from discodop._grammar cimport Grammar
from discodop.containers cimport ChartItem, SmallChartItem, FatChartItem, \
		CFGChartItem, Edge, LCFRSEdge, CFGEdge, RankedEdge, RankedCFGEdge, \
		Rule, LexicalRule, UChar, UInt, ULong, ULLong, logprobadd, logprobsum
cimport cython

from libc.string cimport memset
cdef extern from "macros.h":
	void SETBIT(ULong a[], int b)

REMOVEIDS = re.compile('@[-0-9]+')
BREMOVEIDS = re.compile(b'@[-0-9]+')
REMOVEWORDTAGS = re.compile('@[^ )]+')


cpdef marginalize(method, chart, ChartItem start, Grammar grammar, int n,
		bint sample=False, bint kbest=True, list sent=None, list tags=None,
		int sldop_n=7, list backtransform=None,
		bint bitpar=False):
	""" Approximate MPP or MPD by summing over n random/best derivations from
	chart, return a dictionary mapping parsetrees to probabilities """
	cdef bint mpd = method == 'mpd'
	cdef bint shortest = method == 'shortest'
	cdef Entry entry
	cdef LexicalRule lexrule
	cdef dict parsetrees = {}, derivs = {}
	cdef list derivations = [], entries = []
	cdef str treestr, deriv
	cdef double prob, maxprob
	cdef int m

	assert kbest or sample
	if kbest and not bitpar:
		derivations, D, _ = lazykbest(chart, start, n, grammar.tolabel,
				None, derivs=backtransform is None)
		if isinstance(start, CFGChartItem):
			entries = D[(<CFGChartItem>start).start][
					(<CFGChartItem>start).end][start.label]
		else:
			entries = D[start]
	elif bitpar:
		derivations = sorted(chart.items(), key=itemgetter(1))
		entries = [None] * len(derivations)
		D = {}
	elif isinstance(start, CFGChartItem):
		D = [[{} for right in left] for left in chart]
	else:
		D = {}
	if sample:
		if bitpar:
			return parsetrees, derivs, 'sampling not possible with bitpar.'
		derivations.extend(
				getsamples(D, chart, start, n, grammar.tolabel, None))
		# filter out duplicate derivations
		if isinstance(start, CFGChartItem):
			entries = D[(<CFGChartItem>start).start][
					(<CFGChartItem>start).end][start.label]
		else:
			entries = D[start]
		filteredderivations = dict(zip(derivations, entries))
		entries[:] = filteredderivations.values()
		derivations[:] = filteredderivations.keys()
	if method == 'sl-dop':
		return sldop(dict(derivations), chart, sent, tags, grammar,
				n, sldop_n, backtransform, D, entries, bitpar)
	elif method == 'sl-dop-simple':
		return sldop_simple(dict(derivations), entries, n, sldop_n,
				D, chart, start, grammar, backtransform, bitpar)
	elif method == 'shortest':
		# filter out all derivations which are not shortest
		if backtransform is not None and not bitpar:
			maxprob = min([entry.value for entry in entries])
			entries = [entry for entry in entries if entry.value == maxprob]
		elif derivations:
			_, maxprob = min(derivations, key=itemgetter(1))
			derivations = [(deriv, prob) for deriv, prob in derivations
					if prob == maxprob]

	if backtransform is not None and not bitpar:  # Double-DOP
		for entry in entries:
			prob = entry.value
			treestr = recoverfragments(entry.key, D,
					grammar, backtransform)
			if shortest:
				newprob = exp(-getderivprob(entry.key, D, sent, grammar))
				score = (prob / log(0.5), newprob)
				if treestr not in parsetrees or score > parsetrees[treestr]:
					parsetrees[treestr] = score
					derivs[treestr] = extractfragments(
							entry.key, D, grammar, backtransform)
			elif not mpd and treestr in parsetrees:
				parsetrees[treestr].append(-prob)
			elif not mpd or (treestr not in parsetrees
						or -prob > parsetrees[treestr][0]):
				parsetrees[treestr] = [-prob]
				derivs[treestr] = extractfragments(
						entry.key, D, grammar, backtransform)
	else:  # DOP reduction / bitpar
		for (deriv, prob), entry in zip(derivations, entries):
			if backtransform is None:
				treestr = REMOVEIDS.sub('', deriv)
			else:
				treestr = recoverfragments(deriv, D, grammar, backtransform)
			if shortest:
				# for purposes of tie breaking, calculate the derivation
				# probability in a different model.
				if bitpar:
					# because with bitpar we don't know which rules have been
					# used, read off the rules from the derivation ...
					tree = canonicalize(Tree.parse(deriv, parse_leaf=int))
					newprob = 0.0
					for t in tree.subtrees():
						if isinstance(t[0], Tree):
							if len(t) == 1:
								r = (b'0', t.label, t[0].label)
							elif len(t) == 2:
								r = (b'01', t.label, t[0].label, t[1].label)
							m = grammar.rulenos[r]
							newprob += grammar.bylhs[0][m].prob
						else:
							m = grammar.toid[t.label]
							try:  # FIXME: bitpar smooths tags w/weight < 0.1!
								lexrule = grammar.lexicalbylhs[m][sent[t[0]]]
							except KeyError:
								newprob += 30.0
							else:
								newprob += lexrule.prob
				else:
					newprob = getderivprob(entry.key, D, sent, grammar)
				score = (prob / log(0.5), exp(-newprob))
				if treestr not in parsetrees or score > parsetrees[treestr]:
					parsetrees[treestr] = score
			elif not mpd and treestr in parsetrees:
				# simple way of adding probabilities (too easy):
				parsetrees[treestr].append(-prob)
			elif not mpd or (treestr not in parsetrees
						or -prob > parsetrees[treestr][0]):
				parsetrees[treestr] = [-prob]

	for treestr, probs in parsetrees.items() if not shortest else ():
		parsetrees[treestr] = logprobsum(probs)
	msg = '%d derivations, %d parsetrees' % (
			len(derivations if backtransform is None else entries),
			len(parsetrees))
	return parsetrees, derivs, msg


cdef sldop(dict derivations, chart, list sent, list tags, Grammar grammar,
		int m, int sldop_n, list backtransform, D, entries, bint bitpar):
	""" 'Proper' method for sl-dop. Parses sentence once more to find shortest
	derivations, pruning away any chart item not occurring in the n most
	probable parse trees; we need to parse again because we have to consider
	all derivations for the n most likely trees.

	:returns: the intersection of the most probable parse trees and their
		shortest derivations, with probabilities of the form (subtrees, prob).

	NB: doesn't seem to work so well, so may contain a subtle bug.
		does not support PCFG charts. """
	cdef dict derivs = {}
	# collect derivations for each parse tree
	derivsfortree = defaultdict(set)
	if backtransform is None:
		for deriv in derivations:
			derivsfortree[REMOVEIDS.sub('', deriv)].add(deriv)
	elif bitpar:
		for deriv in derivations:
			derivsfortree[recoverfragments(deriv, D,
					grammar, backtransform)].add(deriv)
	else:
		for entry in entries:
			deriv = getderiv((<Entry>entry).key, D, chart,
					grammar.tolabel, None)
			derivations[deriv] = (<Entry>entry).value
			derivsfortree[recoverfragments((<Entry>entry).key, D,
					grammar, backtransform)].add(deriv)
	# sum over probs of derivations to get probs of parse trees
	parsetreeprob = {tree: logprobsum([-derivations[d] for d in ds])
			for tree, ds in derivsfortree.items()}

	nmostlikelytrees = set(nlargest(sldop_n, parsetreeprob,
			key=parsetreeprob.get))
	grammar.switch(u'shortest', True)
	shortestderivations, DD, msg, start = treeparsing(
			nmostlikelytrees, sent, grammar, m, backtransform, tags)
	if not DD.get(start):
		return {}, {}, msg
	result = {}
	for (deriv, s), entry in zip(shortestderivations, DD[start]):
		if backtransform is None:
			treestr = REMOVEIDS.sub('', deriv)
		else:
			treestr = recoverfragments(deriv if bitpar else (<Entry>entry).key,
					DD, grammar, backtransform)
		if treestr in nmostlikelytrees and treestr not in result:
			result[treestr] = (s / log(0.5), parsetreeprob[treestr])
			if backtransform is not None:
				derivs[treestr] = extractfragments(
						deriv if bitpar else (<Entry>entry).key,
						DD, grammar, backtransform)
			if len(result) > sldop_n:
				break
	if not len(result):
		return {}, {}, 'no matching derivation found'
	msg = '(%d derivations, %d of %d parsetrees)' % (
		len(derivations), min(sldop_n, len(parsetreeprob)), len(parsetreeprob))
	return result, derivs, msg


cdef sldop_simple(dict derivations, list entries, int m, int sldop_n,
		D, chart, start, Grammar grammar, list backtransform, bint bitpar):
	""" simple sl-dop method; estimates shortest derivation directly from
	number of addressed nodes in the k-best derivations. After selecting the n
	best parse trees, the one with the shortest derivation is returned.
	In other words, selects shortest derivation among the list of available
	derivations, instead of finding the shortest among all possible derivations
	using Viterbi. """
	cdef Entry entry
	cdef dict derivs = {}, keys = {}
	derivsfortree = defaultdict(set)
	# collect derivations for each parse tree
	if backtransform is None:
		for deriv in derivations:
			tree = REMOVEIDS.sub('', deriv)
			derivsfortree[tree].add(deriv)
	elif bitpar:
		for deriv in derivations:
			tree = recoverfragments(deriv, D, grammar, backtransform)
			derivsfortree[tree].add(deriv)
	else:
		for entry in entries:
			deriv = getderiv(entry.key, D, chart, grammar.tolabel, '}<')
			tree = recoverfragments(entry.key, D, grammar, backtransform)
			keys[deriv] = entry.key
			derivations[deriv] = entry.value
			derivsfortree[tree].add(deriv)

	# sum over derivations to get parse trees
	parsetreeprob = {tree: logprobsum([-derivations[d] for d in ds])
			for tree, ds in derivsfortree.items()}
	selectedtrees = nlargest(sldop_n, parsetreeprob, key=parsetreeprob.get)

	# the number of fragments used is the number of
	# nodes (open parens), minus the number of interior
	# (addressed) nodes.
	result = {}
	for tree in selectedtrees:
		score, deriv = min([(deriv.count('(') -
				len([a for a in deriv.split() if '@' in a or '}<' in a]),
				deriv)
				for deriv in derivsfortree[tree]])
		result[tree] = (-score, parsetreeprob[tree])
		if backtransform is not None:
			derivs[tree] = extractfragments(deriv if bitpar else keys[deriv],
					D, grammar, backtransform)
	msg = '(%d derivations, %d of %d parsetrees)' % (
			len(derivations), len(result), len(parsetreeprob))
	return result, derivs, msg


def getsamples(D, chart, start, n, tolabel, debin=None):
	""" Samples n derivations from a chart. """
	if isinstance(start, CFGChartItem):
		return getsamples_cfg(D, chart, start, n, tolabel, debin)
	return getsamples_lcfrs(D, chart, start, n, tolabel, debin)


def getsamples_lcfrs(D, chart, start, n, tolabel, debin=None):
	cdef Edge edge
	cdef dict tables = {}
	chartcopy = {}
	for item in chart:
		#FIXME: work w/inside prob right?
		# sort so that highest probability edges come first
		chartcopy[item] = sorted(chart[item])
	for item in chartcopy:
		tables[item] = []
		prev = 0.0
		for edge in chartcopy[item]:
			prev += exp(-edge.inside)
			tables[item].append(prev)
	result = []
	for _ in range(n):
		s, p = samplechart_lcfrs(
				D, chartcopy, start, tolabel, tables, debin)
		result.append((str(s), p))
	return result


cdef samplechart_lcfrs(dict D, dict chart, ChartItem start, list tolabel,
		dict tables, bytes debin):
	""" Samples a derivation from a chart. """
	cdef LCFRSEdge edge
	cdef ChartItem child
	cdef double prob
	lst = tables[start]
	rnd = random() * lst[len(lst) - 1]
	idx = bisect_right(lst, rnd)
	edge = chart[start][idx]
	if edge.left.label == 0:  # == "Epsilon":
		idx = edge.left.lexidx()
		newedge = RankedEdge(start, edge, 0, -1)
		D.setdefault(start, []).append(new_Entry(newedge, edge.inside, 0))
		deriv = "(%s %d)" % (tolabel[start.label].decode('ascii'), idx)
		return deriv, edge.inside
	children = [samplechart_lcfrs(D, chart, child, tolabel, tables, debin)
			for child in (edge.left, edge.right) if child.label]
	if debin is not None and debin in tolabel[start.label]:
		tree = ' '.join([a for a, _ in children])
	else:
		tree = '(%s %s)' % (tolabel[start.label].decode('ascii'),
				' '.join([a for a, _ in children]))
	# create an edge that has as children the edges that were just created
	# by our recursive calls
	newedge = RankedEdge(start, edge, len(D[edge.left]) - 1,
			(len(D[edge.right]) - 1) if edge.right.label else -1)
	prob = edge.rule.prob + fsum([b for _, b in children])
	D.setdefault(start, []).append(new_Entry(newedge, prob, 0))
	return tree, prob


def getsamples_cfg(list D, list chart, CFGChartItem start, int n, list tolabel,
		debin=None):
	cdef Edge edge
	cdef dict tables = {}
	chartcopy = [[{} for r in l] for l in chart]
	for l in range(start.end):
		for r in range(l + 1, start.end + 1):
			for x in chart[l][r]:
				chartcopy[l][r][x] = sorted(chart[l][r][x])
				item = l, r, x
				tables[item] = []
				prev = 0.0
				for edge in chartcopy[l][r][x]:
					prev += exp(-edge.inside)
					tables[item].append(prev)
	result = []
	for _ in range(n):
		s, p = samplechart_cfg(
				D, chartcopy, (start.start, start.end, start.label),
						tolabel, tables, debin)
		result.append((str(s), p))
	return result


cdef samplechart_cfg(list D, list chart, tuple start, list tolabel,
		dict tables, bytes debin):
	""" Samples a derivation from a chart. """
	cdef CFGEdge edge
	cdef double prob
	left, right, label = start
	lst = tables[left, right, label]
	rnd = random() * lst[len(lst) - 1]
	idx = bisect_right(lst, rnd)
	edge = chart[left][right][label][idx]
	if edge.rule is NULL:  # == "Epsilon":
		newedge = RankedCFGEdge(label, left, right, edge, 0, -1)
		D[left][right].setdefault(label, []).append(
				new_Entry(newedge, edge.inside, 0))
		deriv = "(%s %d)" % (tolabel[label].decode('ascii'), left)
		return deriv, edge.inside
	children = [samplechart_cfg(D, chart, child, tolabel, tables, debin)
			for child in ((left, edge.mid, edge.rule.rhs1),
				(edge.mid, right, edge.rule.rhs2)) if child[2]]
	if debin is not None and debin in tolabel[label]:
		tree = ' '.join([a for a, _ in children])
	else:
		tree = '(%s %s)' % (tolabel[label].decode('ascii'),
				' '.join([a for a, _ in children]))
	# create an edge that has as children the edges that were just created
	# by our recursive calls
	newedge = RankedCFGEdge(label, left, right, edge,
			len(D[left][edge.mid][edge.rule.rhs1]) - 1,
			(len(D[edge.mid][right][edge.rule.rhs2]) - 1)
				if edge.rule.rhs2 else -1)
	prob = edge.rule.prob + fsum([b for _, b in children])
	D[left][right].setdefault(label, []).append(new_Entry(newedge, prob, 0))
	return tree, prob


def treeparsing(trees, sent, Grammar grammar, int m, backtransform, tags=None):
	""" Given a sequence of trees (as strings), parse them with a DOP grammar
	to get parse tree probabilities; i.e., will consider multiple derivations.
	"""
	# Parsing & pruning inside the disambiguation module is rather kludgy,
	# but the problem is that we need to get probabilities of trees,
	# not just of derivations. Therefore the coarse-to-fine methods
	# do not apply directly.
	cdef ChartItem item
	cdef FatChartItem fitem
	cdef int x, lensent
	whitelist = [{} for _ in grammar.toid]
	for treestr in trees:
		tree = Tree.parse(treestr, parse_leaf=int)
		lensent = len(tree.leaves())
		for n in tree.subtrees():
			leaves = n.leaves()
			if lensent < sizeof(ULLong) * 8:
				item = SmallChartItem(0, sum([1L << x for x in leaves]))
			else:
				fitem = item = FatChartItem(0)
				for x in leaves:
					SETBIT(fitem.vec, x)
			whitelist[grammar.toid[n.label.encode('ascii')]][item] = 0.0

	# Project labels to all possible labels that generate that label. For DOP
	# reduction, all possible ids; for Double DOP, ignore artificial labels.
	for label, n in grammar.toid.items():
		if backtransform is None:
			whitelist[n] = whitelist[grammar.toid[BREMOVEIDS.sub(b'', label)]]
		elif b'@' in label or b'}<' in label:
			whitelist[n] = None  # do not prune item
		else:
			whitelist[n] = whitelist[grammar.toid[label]]

	# finally, we parse with the small set of allowed labeled spans.
	# we do parse with PCFG even if possible, because that requires a different
	# way of pruning.
	chart, start, _ = plcfrs.parse(sent, grammar, tags=tags,
			whitelist=whitelist)
	if not start:
		return [], {}, "tree parsing failed", None
	return lazykbest(chart, start, m, grammar.tolabel) + (start, )


cdef double getderivprob(deriv, D, sent, Grammar grammar):
	""" Given a derivation as a ranked edge, recursively calculate its
	probability according to a grammar, which has to have matching rules & rule
	numbers. """
	if isinstance(deriv, RankedEdge):
		return getderivprob_lcfrs(deriv, D, sent, grammar)
	elif isinstance(deriv, RankedCFGEdge):
		return getderivprob_cfg(deriv, D, sent, grammar)


cdef double getderivprob_lcfrs(RankedEdge deriv, dict D,
		list sent, Grammar grammar):
	cdef double result
	if deriv.edge.rule is NULL:  # is terminal
		word = sent[deriv.edge.left.lexidx()]
		return (<LexicalRule>grammar.lexicalbylhs[deriv.head.label][word]).prob
	result = grammar.bylhs[0][deriv.edge.rule.no].prob
	result += getderivprob((<Entry>D[deriv.edge.left][deriv.left]).key,
			D, sent, grammar)
	if deriv.edge.right:
		result += getderivprob((<Entry>D[deriv.edge.right][deriv.right]).key,
				D, sent, grammar)
	return result


cdef double getderivprob_cfg(RankedCFGEdge deriv, list D,
		list sent, Grammar grammar):
	cdef double result
	if deriv.edge.rule is NULL:  # is terminal
		word = sent[deriv.start]
		return (<LexicalRule>grammar.lexicalbylhs[deriv.label][word]).prob
	result = grammar.bylhs[0][deriv.edge.rule.no].prob
	result += getderivprob((<Entry>D[deriv.start][deriv.edge.mid][
			deriv.edge.rule.rhs1][deriv.left]).key, D, sent, grammar)
	if deriv.edge.rule.rhs2:
		result += getderivprob((<Entry>D[deriv.edge.mid][deriv.end][
				deriv.edge.rule.rhs2][deriv.right]).key, D, sent, grammar)
	return result


cpdef viterbiderivation(chart, ChartItem start, list tolabel):
	# Ask for at least 10 derivations because unary cycles.
	derivations = lazykbest(chart, start, 10, tolabel)[0]
	return derivations[0]


cpdef str recoverfragments(deriv, D, Grammar grammar, list backtransform):
	""" Reconstruct a DOP derivation from a DOP derivation with
	flattened fragments which are left-binarized. `derivation` should be
	a RankedEdge representing a derivation, and backtransform should contain
	rule numbers as keys and strings as values. Uses the first binarized
	production as key, which map to string templates as values.

	:returns: expanded derivation as a string.

	Does on-the-fly debinarization following labels that are not mapped to a
	label in the coarse grammar, i.e., it assumes that neverblockre is only
	used to avoid blocking nonterminals from the double-dop binarization
	(containing the string '}<'). Note that this means getmapping() has to have
	been called on `grammar`, even when not doing coarse-to-fine parsing. """
	if isinstance(deriv, RankedEdge):
		result = recoverfragments_lcfrs(deriv, D, grammar, backtransform)
	elif isinstance(deriv, RankedCFGEdge):
		result = recoverfragments_cfg(deriv, D, grammar, backtransform)
	elif isinstance(deriv, basestring):
		deriv = Tree.parse(deriv, parse_leaf=int)
		result = recoverfragments_str(deriv, grammar, backtransform)
	else:
		raise ValueError
	return REMOVEWORDTAGS.sub('', result)


cdef str recoverfragments_lcfrs(RankedEdge deriv, dict D,
		Grammar grammar, list backtransform):
	cdef RankedEdge child
	cdef list children = []
	cdef str frag = backtransform[deriv.edge.rule.no]  # template

	# collect all children w/on the fly left-factored debinarization
	if deriv.edge.right.label:  # is there a right child?
		# keep going while left child is part of same binarized constituent
		# instead of looking for a binarization marker in the label string, we
		# use the fact that such labels do not have a mapping as proxy.
		while grammar.mapping[deriv.edge.left.label] == 0:
			# one of the right children
			children.append((<Entry>D[deriv.edge.right][deriv.right]).key)
			# move on to next node in this binarized constituent
			deriv = (<Entry>D[deriv.edge.left][deriv.left]).key
		# last right child
		if deriv.edge.right.label:  # is there a right child?
			children.append((<Entry>D[deriv.edge.right][deriv.right]).key)
	elif grammar.mapping[deriv.edge.left.label] == 0:
		deriv = (<Entry>D[deriv.edge.left][deriv.left]).key
	# left-most child
	children.append((<Entry>D[deriv.edge.left][deriv.left]).key)

	# recursively expand all substitution sites
	# FIXME: to avoid using str + decoding, we could use
	# PyObject* PyBytes_FromFormat(const char *format, ...)
	# PyBytes_FromFormat('(%s %d)', <char *>..., ...)
	children = [('(%s %d)' % (
		str(grammar.tolabel[child.head.label].decode('ascii')),
		child.edge.left.lexidx()))
		if child.edge.rule is NULL else recoverfragments_lcfrs(
				child, D, grammar, backtransform)
				for child in reversed(children)]

	# substitute results in template
	return frag.format(*children)

	# even better: build result incrementally; use bytearray,
	# extended in recursive calls w/strings from backtransform.
	# step 1: collect RankedEdges in a list (children);
	#		i.e., exctract nodes from binarized constituent.
	# step 2: iterate over parts of template, alternately adding string from it
	#		and making a recursive call to insert the relevant child RankedEdge
	# new backtransform format:
	#backtransform[prod] = (list_of_strs, list_of_idx)
	#backtransform[34] = ([b'(NP (DT ', b') (NN ', b'))'], [0, 1])
	#alternatively: (better locality?)
	#frag = backtransform[34] = [b'(NP (DT ', 0, b') (NN ', 1, b'))']
	#result += frag[0]
	#for n in range(1, len(result), 2):
	#	foo(result, children[frag[n]])
	#	result += frag[n + 1]


cdef str recoverfragments_cfg(RankedCFGEdge deriv, list D,
		Grammar grammar, list backtransform):
	cdef RankedCFGEdge child
	cdef list children = []
	cdef str frag = backtransform[deriv.edge.rule.no]  # template

	# collect children w/on the fly left-factored debinarization
	if deriv.edge.rule.rhs2:  # is there a right child?
		# keep going while left child is part of same binarized constituent
		# this shortcut assumes that neverblockre is only used to avoid
		# blocking nonterminals from the double-dop binarization.
		while grammar.mapping[deriv.edge.rule.rhs1] == 0:
			# one of the right children
			children.append((<Entry>D[deriv.edge.mid][deriv.end][
					deriv.edge.rule.rhs2][deriv.right]).key)
			# move on to next node in this binarized constituent
			deriv = (<Entry>D[deriv.start][deriv.edge.mid][
					deriv.edge.rule.rhs1][deriv.left]).key
		# last right child
		if deriv.edge.rule.rhs2:  # is there a right child?
			children.append((<Entry>D[deriv.edge.mid][deriv.end][
					deriv.edge.rule.rhs2][deriv.right]).key)
	elif grammar.mapping[deriv.edge.rule.rhs1] == 0:
		deriv = (<Entry>D[deriv.start][deriv.edge.mid][
				deriv.edge.rule.rhs1][deriv.left]).key
	# left-most child
	children.append((<Entry>D[deriv.start][deriv.edge.mid][
			deriv.edge.rule.rhs1][deriv.left]).key)

	# recursively expand all substitution sites
	children = [('(%s %d)' % (
			str(grammar.tolabel[child.label].decode('ascii')), child.start))
			if child.edge.rule is NULL
			else recoverfragments_cfg(child, D, grammar, backtransform)
			for child in reversed(children)]

	# substitute results in template
	return frag.format(*children)


cdef str recoverfragments_str(deriv, Grammar grammar, list backtransform):
	cdef list children = []
	cdef str frag
	if len(deriv) == 1:
		prod = (b'0', deriv.label, deriv[0].label)
	elif len(deriv) == 2:
		prod = (b'01', deriv.label, deriv[0].label, deriv[1].label)
	frag = backtransform[grammar.rulenos[prod]]  # template
	# collect children w/on the fly left-factored debinarization
	if len(deriv) == 2:  # is there a right child?
		# keep going while left child is part of same binarized constituent
		# this shortcut assumes that neverblockre is only used to avoid
		# blocking nonterminals from the double-dop binarization.
		while '}<' in deriv[0].label:
			# one of the right children
			children.append(deriv[1])
			# move on to next node in this binarized constituent
			deriv = deriv[0]
		# last right child
		if len(deriv) == 2:  # is there a right child?
			children.append(deriv[1])
	elif '}<' in deriv[0].label:
		deriv = deriv[0]
	# left-most child
	children.append(deriv[0])

	# recursively expand all substitution sites
	children = [recoverfragments_str(child, grammar, backtransform)
			if isinstance(child[0], Tree)
			else ('(%s %d)' % (child.label, child[0]))
			for child in reversed(children)]

	# substitute results in template
	return frag.format(*children)


def extractfragments(deriv, D, Grammar grammar, list backtransform):
	result = []
	if isinstance(deriv, RankedEdge):
		extractfragments_lcfrs(deriv, D, grammar, backtransform, result)
	elif isinstance(deriv, RankedCFGEdge):
		extractfragments_cfg(deriv, D, grammar, backtransform, result)
	elif isinstance(deriv, basestring):
		extractfragments_str(deriv, grammar, backtransform, result)
	else:
		raise ValueError
	return result


cdef extractfragments_lcfrs(RankedEdge deriv, dict D,
		Grammar grammar, list backtransform, list result):
	cdef RankedEdge child
	cdef list children = [], labels = []
	cdef str frag = backtransform[deriv.edge.rule.no]  # template

	# collect all children w/on the fly left-factored debinarization
	if deriv.edge.right.label:  # is there a right child?
		# keep going while left child is part of same binarized constituent
		# instead of looking for a binarization marker in the label string, we
		# use the fact that such labels do not have a mapping as proxy.
		while grammar.mapping[deriv.edge.left.label] == 0:
			# one of the right children
			children.append((<Entry>D[deriv.edge.right][deriv.right]).key)
			labels.append(grammar.tolabel[deriv.edge.right.label])
			# move on to next node in this binarized constituent
			deriv = (<Entry>D[deriv.edge.left][deriv.left]).key
		# last right child
		if deriv.edge.right.label:  # is there a right child?
			children.append((<Entry>D[deriv.edge.right][deriv.right]).key)
			labels.append(grammar.tolabel[deriv.edge.right.label])
	elif grammar.mapping[deriv.edge.left.label] == 0:
		deriv = (<Entry>D[deriv.edge.left][deriv.left]).key
	# left-most child
	children.append((<Entry>D[deriv.edge.left][deriv.left]).key)
	labels.append(grammar.tolabel[deriv.edge.left.label])

	frag = frag.format(*['(%s %d)' % (a.split('@')[0], n)
			for n, a in enumerate(reversed(labels))])
	sent = [a[a.index('@') + 1:].decode('unicode-escape') if '@' in a else None
			for a in reversed(labels)]
	result.append((frag, sent))
	# recursively visit all substitution sites
	for child in reversed(children):
		if not child.edge.rule is NULL:
			extractfragments_lcfrs(child, D, grammar, backtransform, result)


cdef extractfragments_cfg(RankedCFGEdge deriv, list D,
		Grammar grammar, list backtransform, list result):
	cdef RankedCFGEdge child
	cdef list children = [], labels = []
	cdef str frag = backtransform[deriv.edge.rule.no]  # template

	# collect all children w/on the fly left-factored debinarization
	if deriv.edge.rule.rhs2:  # is there a right child?
		# keep going while left child is part of same binarized constituent
		# instead of looking for a binarization marker in the label string, we
		# use the fact that such labels do not have a mapping as proxy.
		while grammar.mapping[deriv.edge.rule.rhs1] == 0:
			# one of the right children
			children.append((<Entry>D[deriv.edge.mid][deriv.end][
					deriv.edge.rule.rhs2][deriv.right]).key)
			labels.append(grammar.tolabel[deriv.edge.rule.rhs2])
			# move on to next node in this binarized constituent
			deriv = (<Entry>D[deriv.start][deriv.edge.mid][
					deriv.edge.rule.rhs1][deriv.left]).key
		# last right child
		if deriv.edge.rule.rhs2:  # is there a right child?
			children.append((<Entry>D[deriv.edge.mid][deriv.end][
					deriv.edge.rule.rhs2][deriv.right]).key)
			labels.append(grammar.tolabel[deriv.edge.rule.rhs2])
	elif grammar.mapping[deriv.edge.rule.rhs1] == 0:
		deriv = (<Entry>D[deriv.start][deriv.edge.mid][
				deriv.edge.rule.rhs1][deriv.left]).key
	# left-most child
	children.append((<Entry>D[deriv.start][deriv.edge.mid][
			deriv.edge.rule.rhs1][deriv.left]).key)
	labels.append(grammar.tolabel[deriv.edge.rule.rhs1])

	frag = frag.format(*['(%s %d)' % (a.split('@')[0], n)
			for n, a in enumerate(reversed(labels))])
	sent = [a[a.index('@') + 1:].decode('unicode-escape') if '@' in a else None
			for a in reversed(labels)]
	result.append((frag, sent))
	# recursively visit all substitution sites
	for child in reversed(children):
		if not child.edge.rule is NULL:
			extractfragments_cfg(child, D, grammar, backtransform, result)


cdef str extractfragments_str(deriv, Grammar grammar,
		list backtransform, list result):
	cdef list children = [], labels = []
	cdef str frag
	if len(deriv) == 1:
		prod = (b'0', deriv.label, deriv[0].label)
	elif len(deriv) == 2:
		prod = (b'01', deriv.label, deriv[0].label, deriv[1].label)
	frag = backtransform[grammar.rulenos[prod]]  # template
	# collect children w/on the fly left-factored debinarization
	if len(deriv) == 2:  # is there a right child?
		# keep going while left child is part of same binarized constituent
		# this shortcut assumes that neverblockre is only used to avoid
		# blocking nonterminals from the double-dop binarization.
		while '}<' in deriv[0].label:
			# one of the right children
			children.append(deriv[1])
			labels.append(deriv[1].label)
			# move on to next node in this binarized constituent
			deriv = deriv[0]
		# last right child
		if len(deriv) == 2:  # is there a right child?
			children.append(deriv[1])
			labels.append(deriv[1].label)
	elif '}<' in deriv[0].label:
		deriv = deriv[0]
	# left-most child
	children.append(deriv[0])
	labels.append(deriv[0].label)

	frag = frag.format(*['(%s %d)' % (a.split('@')[0], n)
			for n, a in enumerate(reversed(labels))])
	sent = [a[a.index('@') + 1:].decode('unicode-escape') if '@' in a else None
			for a in reversed(labels)]
	result.append((frag, sent))
	# recursively visit all substitution sites
	for child in reversed(children):
		if isinstance(child[0], Tree):
			extractfragments_str(child, grammar, backtransform, result)


def doprerank(chart, start, sent, n, Grammar coarse, Grammar fine):
	""" Given a chart from a coarse stage, re-rank its n-best derivations with
	DOP parse probabilities of a DOP reduction (cf. ``dopparseprob()``). """
	cdef dict results = {}
	derivations, _, _ = lazykbest(chart, start, n, coarse.tolabel,
			None, derivs=True)
	for derivstr, _ in derivations:
		deriv = addbitsets(derivstr)
		results[derivstr] = exp(dopparseprob(deriv, sent, coarse, fine))
	return results


def dopparseprob(tree, sent, Grammar coarse, Grammar fine):
	""" Given a Tree and a DOP reduction, compute the exact DOP parse
	probability.

	This follows up on a suggestion made by Goodman (2003, p. 143)
	of calculating DOP probabilities of given parse trees, although I'm not
	sure it has complexity O(nP) as he suggests (with n as number of nodes in
	input, and P as max number of rules consistent with a node in the input).
	Furthermore, the idea of sampling trees "long enough" until we have the MPP
	is no faster than sampling without applying this procedure, because to
	determine that some probability p is the maximal probability, we need to
	collect the probability mass p_seen of enough parse trees such that we have
	some parsetree with probability p > (1 - p_seen), which requires first
	seeing almost all parse trees, unless p is exceptionally high. Hence, this
	method is mostly useful in a reranking framework where it is known in
	advance that a small set of trees is of interest.

	Expects a mapping which gives a list of consistent rules from the reduction
	as produced by ``fine.getrulemapping(coarse)``.

	NB: this algorithm could also be used to determine the probability of
	derivations, but then the input would have to distinguish whether nodes are
	internal nodes of fragments, or whether they join two fragments. """
	neginf = float('-inf')
	cdef dict chart = {}  # chart[label, left, right] = prob
	cdef tuple a, b, c
	cdef Rule *rule
	cdef LexicalRule lexrule
	assert fine.logprob, 'Grammar should have log probabilities.'
	# Log probabilities are not ideal here because we do lots of additions,
	# but the probabilities are very small.
	# A possible alternative is to scale them somehow.

	# add all matching POS tags
	for n, pos in tree.pos():
		pos = pos.encode('ascii')
		word = sent[n]
		for lexrule in fine.lexicalbyword[word]:
			if (fine.tolabel[lexrule.lhs] == pos
					or fine.tolabel[lexrule.lhs].startswith(pos + b'@')):
				chart[lexrule.lhs, 1 << n] = logprobadd(
					chart.get((lexrule.lhs, 1 << n), neginf), -lexrule.prob)

	# do post-order traversal (bottom-up)
	for node, (prod, yf) in list(zip(tree.subtrees(),
			lcfrs_productions(tree, sent)))[::-1]:
		if not isinstance(node[0], Tree):
			continue
		yf = ','.join(''.join(map(str, a)) for a in yf)
		prod = coarse.rulenos[(yf, ) + prod]
		if len(node) == 1:  # unary node
			for ruleno in fine.rulemapping[prod]:
				rule = &(fine.bylhs[0][ruleno])
				b = (rule.rhs1, node.bitset)
				if b in chart:
					a = (rule.lhs, node.bitset)
					if a in chart:
						chart[a] = logprobadd(chart[a], -rule.prob + chart[b])
					else:
						chart[a] = (-rule.prob + chart[b])
		elif len(node) == 2:  # binary node
			for ruleno in fine.rulemapping[prod]:
				rule = &(fine.bylhs[0][ruleno])
				b = (rule.rhs1, node[0].bitset)
				c = (rule.rhs2, node[1].bitset)
				if b in chart and c in chart:
					a = (rule.lhs, node.bitset)
					if a in chart:
						chart[a] = logprobadd(chart[a],
							(-rule.prob + chart[b] + chart[c]))
					else:
						chart[a] = -rule.prob + chart[b] + chart[c]
		else:
			raise ValueError('expected binary tree without empty nodes.')
	return chart.get((fine.toid[tree.label], tree.bitset), neginf)


def main():
	from discodop.grammar import dopreduction
	from discodop._grammar import Grammar
	from discodop import plcfrs

	def e(x):
		a, b = x
		if isinstance(b, tuple):
			return (a.replace("@", ''), (int(abs(b[0])), b[1]))
		return a.replace("@", ''), b

	def maxitem(d):
		return max(d.items(), key=itemgetter(1))

	trees = [Tree.parse(t, parse_leaf=int) for t in
		"""(ROOT (A (A 0) (B 1)) (C 2))
		(ROOT (C 0) (A (A 1) (B 2)))
		(ROOT (B (A 0) (B 1)) (C 2))
		(ROOT (B (A 0) (B 1)) (C 2))
		(ROOT (B (A 0) (B 1)) (C 2))
		(ROOT (B (A 0) (B 1)) (C 2))
		(ROOT (B (A 0) (B 1)) (C 2))
		(ROOT (B (A 0) (B 1)) (C 2))
		(ROOT (B (A 0) (B 1)) (C 2))
		(ROOT (B (A 0) (B 1)) (C 2))
		(ROOT (B (A 0) (B 1)) (C 2))
		(ROOT (B (A 0) (B 1)) (C 2))
		(ROOT (B (A 0) (B 1)) (C 2))
		(ROOT (B (A 0) (B 1)) (C 2))
		(ROOT (B (A 0) (B 1)) (C 2))
		(ROOT (B (A 0) (B 1)) (C 2))
		(ROOT (A 0) (C (B 1) (C 2)))
		(ROOT (A 0) (C (B 1) (C 2)))""".splitlines()]
	sents = [a.split() for a in
		"""d b c\n c a b\n a e f\n a e f\n a e f\n a e f\n d b f\n d b f
		d b f\n d b g\n e f c\n e f c\n e f c\n e f c\n e f c\n e f c\n f b c
		a d e""".splitlines()]
	xgrammar, altweights = dopreduction(trees, sents)
	grammar = Grammar(xgrammar)
	grammar.register(u'shortest', altweights['shortest'])
	print(grammar)
	sent = "a b c".split()
	chart, start, _ = plcfrs.parse(sent, grammar, None, True)
	assert start
	vitderiv, vitprob = viterbiderivation(chart, start, grammar.tolabel)
	mpd, _, _ = marginalize("mpd", chart, start, grammar, 1000)
	mpp, _, _ = marginalize("mpp", chart, start, grammar, 1000)
	mppsampled, _, _ = marginalize("mpp", chart, start, grammar, 1000,
			sample=True, kbest=False)
	sldop1, _, _ = marginalize("sl-dop", chart, start, grammar, 1000,
			sldop_n=7, sent=sent)
	sldopsimple, _, _ = marginalize("sl-dop-simple", chart, start, grammar,
			1000, sldop_n=7, sent=sent)
	short, _, _ = marginalize("shortest", chart, start, grammar,
		1000, sent=sent)
	print("\nvit:\t\t%s %r" % e((REMOVEIDS.sub('', vitderiv), exp(-vitprob))),
		"MPD:\t\t%s %r" % e(maxitem(mpd)),
		"MPP:\t\t%s %r" % e(maxitem(mpp)),
		"MPP sampled:\t%s %r" % e(maxitem(mppsampled)),
		"SL-DOP n=7:\t%s %r" % e(maxitem(sldop1)),
		"simple SL-DOP:\t%s %r" % e(maxitem(sldopsimple)),
		"shortest:\t%s %r" % e(maxitem(short)), sep='\n')

if __name__ == '__main__':
	main()



""" Objects for grammars and grammar rules. """
import re
import logging
import numpy as np
from math import fsum, exp
cimport cython

# This regex should match exactly the set of valid yield functions,
# i.e., comma-separated strings of alternating occurrences from the set {0,1},
YFBINARY = re.compile(
		br'^(?:0|1|1?(?:01)+|0?(?:10)+)(?:,(?:0|1|1?(?:01)+|0?(?:10)+))*$')
YFUNARYRE = re.compile(br'^0(?:,0)*$')
# Match when non-integral weights are present
LCFRS_NONINT = re.compile(b"\t[0-9]+[./][0-9]+\n")
BITPAR_NONINT = re.compile(b"(?:^|\n)[0-9]+\.[0-9]+[ \t]")
LEXICON_NONINT = re.compile("[ \t][0-9]+[./][0-9]+[ \t\n]")

# comparison functions for sorting rules on LHS/RHS labels.
cdef int cmp0(const void *p1, const void *p2) nogil:
	cdef Rule *a = <Rule *>p1, *b = <Rule *>p2
	if a.lhs == b.lhs:
		return (a.no > b.no) - (a.no < b.no)
	return (a.lhs > b.lhs) - (a.lhs < b.lhs)
cdef int cmp1(const void *p1, const void *p2) nogil:
	cdef Rule *a = <Rule *>p1, *b = <Rule *>p2
	return (a.rhs1 > b.rhs1) - (a.rhs1 < b.rhs1)
cdef int cmp2(const void *p1, const void *p2) nogil:
	cdef Rule *a = <Rule *>p1, *b = <Rule *>p2
	return (a.rhs2 > b.rhs2) - (a.rhs2 < b.rhs2)


cdef class Grammar:
	""" A grammar object which stores rules compactly, indexed in various ways.

	:param rule_tuples_or_bytes: either a sequence of tuples containing both
		phrasal & lexical rules, or a bytes string containing the phrasal
		rules in text format; in the latter case lexicon should be given.
		The text format allows for more efficient loading and is used
		internally.
	:param start: a string identifying the unique start symbol of this grammar,
		which will be used by default when parsing with this grammar
	:param bitpar: whether to expect and use the bitpar grammar format

	By default the grammar is in logprob mode;
	invoke grammar.switch('default', logprob=False) to switch.
	If the grammar only contains integral weights (frequencies), they will
	be normalized into relative frequencies; if the grammar contains any
	non-integral weights, weights will be left unchanged. """
	def __cinit__(self):
		self.fanout = self.unary = self.mapping = self.splitmapping = NULL

	def __init__(self, rule_tuples_or_bytes, lexicon=None, start=b'ROOT',
			bitpar=False):
		cdef LexicalRule lexrule
		cdef double [:] tmp
		cdef int n
		self.mapping = self.splitmapping = self.bylhs = NULL
		if not isinstance(start, bytes):
			start = start.encode('ascii')
		self.start = start
		self.bitpar = bitpar
		self.numunary = self.numbinary = self.currentmodel = 0
		self.modelnames = [u'default']
		self.logprob = False

		if rule_tuples_or_bytes and isinstance(rule_tuples_or_bytes, bytes):
			assert isinstance(lexicon, unicode), "expected lexicon"
			self.origrules = rule_tuples_or_bytes
			self.origlexicon = lexicon
		elif rule_tuples_or_bytes and isinstance(
				rule_tuples_or_bytes[0], tuple):
			# convert tuples to strings with text format
			from discodop.grammar import write_lcfrs_grammar
			self.origrules, self.origlexicon = write_lcfrs_grammar(
					rule_tuples_or_bytes, bitpar=bitpar)
		else:
			raise ValueError(
					'expected non-empty sequence of tuples or bytes string.')

		# collect non-terminal labels; count number of rules in each category
		# for allocation purposes.
		rulelines = self.origrules.splitlines()
		fanoutdict = self._countrules(rulelines)
		self._convertlexicon(fanoutdict)
		self.tolabel = sorted(self.toid, key=self.toid.get)
		self.nonterminals = len(self.toid)
		self._allocate()
		self._convertrules(rulelines, fanoutdict)
		for n in range(self.nonterminals):
			self.fanout[n] = fanoutdict[self.tolabel[n]]
		del rulelines, fanoutdict
		# index & filter phrasal rules in different ways
		self._indexrules(self.bylhs, 0, 0)
		# if the grammar only contains integral values (frequencies),
		# normalize them into relative frequencies.
		nonint = BITPAR_NONINT if self.bitpar else LCFRS_NONINT
		if not (nonint.search(self.origrules)
				or LEXICON_NONINT.search(self.origlexicon)):
			self._normalize()
		# store 'default' weights
		self.models = np.empty((1, self.numrules + len(self.lexical)), dtype='d')
		tmp = self.models[0]
		for n in range(self.numrules):
			tmp[self.bylhs[0][n].no] = self.bylhs[0][n].prob
		for n, lexrule in enumerate(self.lexical, self.numrules):
			tmp[n] = lexrule.prob
		self._indexrules(self.unary, 1, 2)
		self._indexrules(self.lbinary, 1, 3)
		self._indexrules(self.rbinary, 2, 3)
		self.switch('default', True)  # enable log probabilities
		for n in range(self.numrules):  # sanity check
			assert self.bylhs[0][n].no == n, self.rulestr(self.bylhs[0][n])

	@cython.wraparound(True)
	def _countrules(self, list rulelines):
		""" Count unary & binary rules; make a canonical list of all
		non-terminal labels and assign them unique IDs """
		Epsilon = b'Epsilon'
		# Epsilon gets IDs 0, can only occur in RHS of lexical rules.
		self.toid = {Epsilon: 0}
		fanoutdict = {Epsilon: 0}  # temporary mapping of labels to fan-outs
		for line in rulelines:
			if not line:
				continue
			fields = line.split()
			if self.bitpar:
				rule = fields[1:]
				yf = b'0' if len(rule) == 2 else b'01'
			else:
				rule = fields[:-2]
				yf = fields[-2]
			assert Epsilon not in rule, ("Epsilon symbol is only used "
						"to introduce terminal symbols in lexical rules.")
			assert self.start not in rule[1:], (
					"Start symbol should only occur on LHS.")
			if len(rule) == 2:
				assert YFUNARYRE.match(yf), ("yield function refers to "
						"non-existent second non-terminal: %r\t%r" % (rule, yf))
				self.numunary += 1
			elif len(rule) == 3:
				assert YFBINARY.match(yf), "illegal yield function: %s" % yf
				assert b'0' in yf and b'1' in yf, ("mismatch between "
						"non-terminals and yield function: %r\t%r" % (rule, yf))
				self.numbinary += 1
			else:
				raise ValueError("grammar not binarized:\n%s" % line)
			if rule[0] not in self.toid:
				self.toid[rule[0]] = len(self.toid)
				fanout = yf.count(b',') + 1
				fanoutdict[rule[0]] = fanout
				if fanout > self.maxfanout:
					self.maxfanout = fanout

		assert self.start in self.toid, ("Start symbol %r not in set of "
				"non-terminal labels extracted from grammar rules." % self.start)
		self.numrules = self.numunary + self.numbinary
		assert self.numrules, "no rules found"
		return fanoutdict

	def _convertlexicon(self, fanoutdict):
		""" Make objects for lexical rules. """
		cdef int n, x
		self.lexical = []
		self.lexicalbyword = {}
		self.lexicalbylhs = {}
		for line in self.origlexicon.splitlines():
			if not line:
				continue
			x = line.index('\t')
			word = line[:x]
			fields = line[x + 1:].encode('ascii').split()
			assert word not in self.lexicalbyword, (
					"word %r appears more than once in lexicon file" % word)
			self.lexicalbyword[word] = []
			for tag, w in zip(fields[::2], fields[1::2]):
				if tag not in self.toid:
					self.toid[tag] = len(self.toid)
					fanoutdict[tag] = 1
					#logging.warning("POS tag %r for word %r "
					#		"not used in any phrasal rule", tag, word)
					#continue
				else:
					assert fanoutdict[tag] == 1, (
							"POS tag %r does not have fan-out 1." % tag)
				# convert fraction to float
				x = w.find(b'/')
				w = float(w[:x]) / float(w[x + 1:]) if x > 0 else float(w)
				assert w > 0, (
						"weights should be positive and non-zero:\n%r" % line)
				lexrule = LexicalRule(self.toid[tag], word, w)
				if lexrule.lhs not in self.lexicalbylhs:
					self.lexicalbylhs[lexrule.lhs] = {}
				self.lexical.append(lexrule)
				self.lexicalbyword[word].append(lexrule)
				self.lexicalbylhs[lexrule.lhs][word] = lexrule
			assert self.lexical and self.lexicalbyword and self.lexicalbylhs, (
					"no lexical rules found.")

	def _allocate(self):
		""" Allocate memory to store rules. """
		# store all non-lexical rules in a contiguous array
		# the other arrays will contain pointers to relevant parts thereof
		# (indexed on lhs, rhs1, and rhs2 of rules)
		self.bylhs = <Rule **>malloc(sizeof(Rule *) * self.nonterminals * 4)
		assert self.bylhs is not NULL
		self.bylhs[0] = NULL
		self.unary = &(self.bylhs[1 * self.nonterminals])
		self.lbinary = &(self.bylhs[2 * self.nonterminals])
		self.rbinary = &(self.bylhs[3 * self.nonterminals])
		# allocate the actual contiguous array that will contain the rules
		# (plus sentinels)
		self.bylhs[0] = <Rule *>malloc(sizeof(Rule) *
			(self.numrules + (2 * self.numbinary) + self.numunary + 4))
		assert self.bylhs[0] is not NULL
		self.unary[0] = &(self.bylhs[0][self.numrules + 1])
		self.lbinary[0] = &(self.unary[0][self.numunary + 1])
		self.rbinary[0] = &(self.lbinary[0][self.numbinary + 1])
		self.fanout = <UChar *>malloc(sizeof(UChar) * self.nonterminals)
		assert self.fanout is not NULL

	@cython.wraparound(True)
	cdef _convertrules(Grammar self, list rulelines, dict fanoutdict):
		""" Auxiliary function to create Grammar objects. Copies grammar
		rules from a text file to a contiguous array of structs. """
		cdef UInt n = 0, m, prev = self.nonterminals
		cdef Rule *cur
		self.rulenos = {}
		for line in rulelines:
			if not line:
				continue
			fields = line.split()
			if self.bitpar:
				rule = fields[1:]
				yf = b'0' if len(rule) == 2 else b'01'
				w = fields[0]
			else:
				rule = fields[:-2]
				yf = fields[-2]
				w = fields[-1]
			# check whether RHS labels have been seen as LHS and check fanout
			for m, nt in enumerate(rule):
				assert nt in self.toid, ('symbol %r has not been seen as LHS '
						'in any rule: %s' % (nt, line))
				fanout = yf.count(b',01'[m:m + 1]) + (m == 0)
				assert fanoutdict[nt] == fanout, (
						"conflicting fanouts for symbol '%s'.\n"
						"previous: %d; this non-terminal: %d.\nrule: %r" % (
						nt, fanoutdict[nt], fanout, rule))
			# convert fraction to float
			x = w.find(b'/')
			w = float(w[:x]) / float(w[x + 1:]) if x > 0 else float(w)
			assert w > 0, "weights should be positive and non-zero:\n%r" % line
			# n is the rule index in the array, and will be the ID for the rule
			cur = &(self.bylhs[0][n])
			cur.no = n
			cur.lhs = self.toid[rule[0]]
			cur.rhs1 = self.toid[rule[1]]
			cur.rhs2 = self.toid[rule[2]] if len(rule) == 3 else 0
			cur.prob = w
			cur.lengths = cur.args = m = 0
			for a in yf.decode('ascii'):
				if a == ',':
					cur.lengths |= 1 << (m - 1)
					continue
				elif a == '1':
					cur.args += 1 << m
				elif a != '0':
					raise ValueError('expected: %r; got: %r' % ('0', a))
				m += 1
			cur.lengths |= 1 << (m - 1)
			assert m < (8 * sizeof(cur.args)), (m, (8 * sizeof(cur.args)))
			self.rulenos[(yf, ) + tuple(rule)] = n
			n += 1
		assert n == self.numrules, (n, self.numrules)

	def _normalize(self):
		""" Optionally normalize frequencies to relative frequencies.
		Should be run during initialization. """
		cdef double mass = 0
		cdef UInt n = 0, lhs
		cdef LexicalRule lexrule
		for lhs in range(self.nonterminals):
			mass = 0
			n = 0
			while self.bylhs[lhs][n].lhs == lhs:
				mass += self.bylhs[lhs][n].prob
				n += 1
			for lexrule in self.lexicalbylhs.get(lhs, {}).values():
				mass += lexrule.prob
			n = 0
			while self.bylhs[lhs][n].lhs == lhs:
				self.bylhs[lhs][n].prob /= mass
				n += 1
			for lexrule in self.lexicalbylhs.get(lhs, {}).values():
				lexrule.prob /= mass

	cdef _indexrules(Grammar self, Rule **dest, int idx, int filterlen):
		""" Auxiliary function to create Grammar objects. Copies certain
		grammar rules and sorts them on the given index.
		Resulting array is ordered by lhs, rhs1, or rhs2 depending on the value
		of `idx` (0, 1, or 2); filterlen can be 0, 2, or 3 to get all, only
		unary, or only binary rules, respectively.
		A separate array has a pointer for each non-terminal into this array;
		e.g.: dest[NP][0] == the first rule with an NP in the idx position. """
		cdef UInt prev = self.nonterminals, idxlabel = 0, n, m = 0
		cdef Rule *cur
		#need to set dest even when there are no rules for that idx
		for n in range(self.nonterminals):
			dest[n] = dest[0]
		if dest is self.bylhs:
			m = self.numrules
		else:
			for n in range(self.numrules):
				if (filterlen == 2) == (self.bylhs[0][n].rhs2 == 0):
					# copy this rule
					dest[0][m] = self.bylhs[0][n]
					assert dest[0][m].no < self.numrules
					m += 1
		if filterlen == 2:
			assert m == self.numunary, (m, self.numunary)
		elif filterlen == 3:
			assert m == self.numbinary, (m, self.numbinary)
		# sort rules by idx
		if idx == 0:
			qsort(dest[0], m, sizeof(Rule), &cmp0)
		elif idx == 1:
			qsort(dest[0], m, sizeof(Rule), &cmp1)
		elif idx == 2:
			qsort(dest[0], m, sizeof(Rule), &cmp2)
		# make index: dest[NP] points to first rule with NP in index position
		for n in range(m):
			cur = &(dest[0][n])
			if idx == 0:
				idxlabel = cur.lhs
			elif idx == 1:
				idxlabel = cur.rhs1
			elif idx == 2:
				idxlabel = cur.rhs2
			if idxlabel != prev:
				dest[idxlabel] = cur
			prev = idxlabel
			assert cur.no < self.numrules
		# sentinel rule
		dest[0][m].lhs = dest[0][m].rhs1 = dest[0][m].rhs2 = self.nonterminals

	def register(self, unicode name, weights):
		""" Register a probabilistic model given a name, and sequences of
		weights 'weights', where weights are in the same order as that of
		self.origrules and self.origlexicon (which is an arbitrary order
		except that words appear sorted). """
		cdef int n, m = len(self.modelnames)
		cdef double [:] tmp
		assert name not in self.modelnames, 'model %r already exists' % name
		assert len(self.modelnames) <= 255, (
				'256 probabilistic models should be enough for anyone.')
		assert len(weights) == self.numrules + len(self.lexical), (
				'length mismatch: %d grammar rules, %d weights given.' % (
					self.numrules + len(self.lexical), len(weights)))
		self.models.resize(m + 1, self.numrules + len(self.lexical))
		self.modelnames.append(name)
		tmp = self.models[m]
		for n in range(self.numrules):
			tmp[self.bylhs[0][n].no] = weights[n]
		for n, lexrule in enumerate(self.lexical, self.numrules):
			tmp[n] = weights[n]

	def switch(self, name, bint logprob=True):
		""" Switch to a different probabilistic model;
		use 'default' to swith back to model given during initialization. """
		cdef int n, m = self.modelnames.index(name)
		cdef double [:] tmp
		cdef LexicalRule lexrule
		if self.currentmodel == m and self.logprob == logprob:
			return
		tmp = -np.log(self.models[m]) if logprob else self.models[m]
		for n in range(self.numrules):
			self.bylhs[0][n].prob = tmp[self.bylhs[0][n].no]
		for n in range(self.numbinary):
			self.lbinary[0][n].prob = tmp[self.lbinary[0][n].no]
			self.rbinary[0][n].prob = tmp[self.rbinary[0][n].no]
		for n in range(self.numunary):
			self.unary[0][n].prob = tmp[self.unary[0][n].no]
		for n, lexrule in enumerate(self.lexical, self.numrules):
			lexrule.prob = tmp[n]
		self.logprob = logprob
		self.currentmodel = m

	def buildchainvec(self):
		""" Build a boolean matrix representing the unary (chain) rules. """
		cdef UInt n
		cdef Rule *rule
		self.chainvec = <ULong *>calloc(self.nonterminals
				* BITNSLOTS(self.nonterminals), sizeof(ULong))
		assert self.chainvec is not NULL
		for n in range(self.numunary):
			rule = self.unary[n]
			SETBIT(self.chainvec, rule.rhs1 * self.nonterminals + rule.lhs)

	def testgrammar(self, epsilon=1e-100):
		""" Report whether all left-hand sides sum to 1 +/-epsilon for the
		currently selected weights. """
		cdef Rule *rule
		cdef LexicalRule lexrule
		cdef UInt n
		cdef list sums = [[] for _ in self.toid]
		cdef double [:] tmp = self.models[self.currentmodel, :]
		#We could be strict about separating POS tags and phrasal categories,
		#but Negra contains at least one tag (--) used for both.
		for n in range(self.numrules):
			rule = &(self.bylhs[0][n])
			sums[rule.lhs].append(tmp[rule.no])
		for n, lexrule in enumerate(self.lexical, self.numrules):
			sums[lexrule.lhs].append(tmp[n])
		for lhs, weights in enumerate(sums[1:], 1):
			mass = fsum(weights)
			if abs(mass - 1.0) > epsilon:
				logging.error("Weights of rules with LHS '%s' "
						"do not sum to 1 +/- %g: sum = %g; diff = %g",
						self.tolabel[lhs], epsilon, mass, mass - 1.0)
				return False
		logging.info("All left hand sides sum to 1 +/- epsilon=%s", epsilon)
		return True

	def getmapping(Grammar self, Grammar coarse, striplabelre=None,
			neverblockre=None, bint splitprune=False, bint markorigin=False):
		""" Construct a mapping of fine non-terminal IDs to coarse non-terminal
		IDs, by applying the regex ``striplabelre`` to the fine labels, used
		for coarse-to-fine pruning. A secondary regex neverblockre is for items
		that should never be pruned.
		The regexes should be compiled objects, i.e., ``re.compile(regex)``,
		or ``None`` to leave labels unchanged. ``coarse`` may be ``None``, in
		which case only ``neverblockre`` is effective, which is also used for
		non-pruning purposes.

		- use ``|<`` to ignore nodes introduced by binarization;
			useful if coarse and fine stages employ different kinds of
			markovization; e.g., ``NP`` and ``VP`` may be blocked,
			but not ``NP|<DT-NN>``.
		- ``_[0-9]+`` to ignore discontinuous nodes ``X_n`` where ``X`` is a
			label and *n* is a fanout. """
		cdef int n, m, components = 0
		cdef set seen = {0}
		if coarse is None:
			coarse = self
		if self.mapping is not NULL:
			free(self.mapping)
		self.mapping = <UInt *>malloc(sizeof(UInt) * self.nonterminals)
		if splitprune and markorigin:
			if self.splitmapping is not NULL:
				if self.splitmapping[0] is not NULL:
					free(self.splitmapping[0])
				free(self.splitmapping)
			self.splitmapping = <UInt **>malloc(sizeof(UInt *)
					* self.nonterminals)
			for n in range(self.nonterminals):
				self.splitmapping[n] = NULL
			self.splitmapping[0] = <UInt *>malloc(sizeof(UInt) *
				sum([self.fanout[n] for n in range(self.nonterminals)
					if self.fanout[n] > 1]))
		for n in range(self.nonterminals):
			if not neverblockre or neverblockre.search(self.tolabel[n]) is None:
				strlabel = self.tolabel[n]
				if striplabelre is not None:
					strlabel = striplabelre.sub(b'', strlabel, 1)
				if self.fanout[n] > 1 and splitprune:
					strlabel += b'*'
				if self.fanout[n] > 1 and splitprune and markorigin:
					self.mapping[n] = self.nonterminals  # sentinel value
					self.splitmapping[n] = &(self.splitmapping[0][components])
					components += self.fanout[n]
					for m in range(self.fanout[n]):
						self.splitmapping[n][m] = coarse.toid[
								strlabel + str(m).encode('ascii')]
						seen.add(self.splitmapping[n][m])
				else:
					self.mapping[n] = coarse.toid[strlabel]
					seen.add(self.mapping[n])
			else:
				self.mapping[n] = 0
		if seen == set(range(coarse.nonterminals)):
			msg = 'label sets are equal'
		else:
			# NB: ALL fine symbols are mapped to some coarse symbol;
			# we only check if all coarse symbols have received a mapping.
			l = sorted([coarse.tolabel[a].decode('ascii') for a in
					set(range(coarse.nonterminals)) - seen])
			diff = ", ".join(l[:10]) + (', ...' if len(l) > 10 else '')
			if coarse.nonterminals > self.nonterminals:
				msg = ('grammar is not a superset of coarse grammar:\n'
						'coarse labels without mapping: {%s}' % diff)
			elif coarse.nonterminals < self.nonterminals:
				msg = 'grammar is a proper superset of coarse grammar.'
			else:
				msg = ('equal number of nodes, but not equivalent:\n'
						'coarse labels without mapping: {%s}' % diff)
		return msg

	def getrulemapping(Grammar self, Grammar coarse):
		""" Produce a mapping of coarse rules to sets of fine rules;
		e.g., ``mapping[12] == [34, 56, 78, ...]``.
		For use with ``dopparseprob()``. """
		cdef list mapping = [[] for _ in range(coarse.numrules)]
		cdef Rule *rule
		for ruleno in range(self.numrules):
			rule = &(self.bylhs[0][ruleno])
			key = (self.yfstr(rule[0]),
					self.tolabel[rule.lhs].rsplit('@', 1)[0],
					self.tolabel[rule.rhs1].rsplit('@', 1)[0])
			if rule.rhs2:
				key += (self.tolabel[rule.rhs2].rsplit('@', 1)[0], )
			mapping[coarse.rulenos[key]].append(ruleno)
		self.rulemapping = mapping

	cdef rulestr(self, Rule rule):
		left = '%.2f %s => %s%s' % (
			exp(-rule.prob) if self.logprob else rule.prob,
			self.tolabel[rule.lhs].decode('ascii'),
			self.tolabel[rule.rhs1].decode('ascii'),
			'  %s' % self.tolabel[rule.rhs2].decode('ascii')
				if rule.rhs2 else '')
		return '%s%s [%d]' % (left.ljust(40), self.yfstr(rule), rule.no)

	cdef yfstr(self, Rule rule):
		cdef int n, m = 0
		cdef result = ''
		for n in range(8 * sizeof(rule.args)):
			result += '1' if (rule.args >> n) & 1 else '0'
			if (rule.lengths >> n) & 1:
				m += 1
				if m == self.fanout[rule.lhs]:
					return result
				else:
					result += ","
		raise ValueError("illegal yield function expected %d components.\n"
				"args: %s; lengths: %s" % (self.fanout[rule.lhs],
				bin(rule.args), bin(rule.lengths)))

	def __str__(self):
		cdef LexicalRule lexrule
		cdef size_t n
		rules = "\n".join(filter(None,
			[self.rulestr(self.bylhs[0][n]) for n in range(self.numrules)]))
		lexical = "\n".join(["%.2f %s => %s" % (
				exp(-lexrule.prob) if self.logprob else lexrule.prob,
				self.tolabel[lexrule.lhs].decode('ascii'),
				lexrule.word.encode('unicode-escape').decode('ascii'))
			for word in sorted(self.lexicalbyword)
			for lexrule in sorted(self.lexicalbyword[word],
			key=lambda lexrule: (<LexicalRule>lexrule).lhs)])
		labels = ", ".join("%s=%d" % (a.decode('ascii'), b)
				for a, b in sorted(self.toid.items()))
		return "rules:\n%s\nlexicon:\n%s\nlabels:\n%s" % (
				rules, lexical, labels)

	def __repr__(self):
		return "%s(\n%s,\n%s\n)" % (self.__class__.__name__,
				self.origrules, self.origlexicon)

	def __reduce__(self):
		""" Helper function for pickling. """
		return (Grammar, (self.origrules, self.origlexicon,
				self.start, self.logprob, self.bitpar))

	def __dealloc__(self):
		if self.bylhs is NULL:
			return
		free(self.bylhs[0])
		self.bylhs[0] = NULL
		free(self.bylhs)
		self.bylhs = NULL
		free(self.fanout)
		self.fanout = NULL
		if self.chainvec is not NULL:
			free(self.chainvec)
			self.chainvec = NULL
		if self.mapping is not NULL:
			free(self.mapping)
			self.mapping = NULL
		if self.splitmapping is not NULL:
			free(self.splitmapping[0])
			free(self.splitmapping)
			self.splitmapping = NULL



""" Web interface to search a treebank. Requires Flask, tgrep2, style.
Expects one or more treebanks with .mrg extension in the directory corpus/ """
# stdlib
import os
import re
import cgi
import glob
import logging
import tempfile
import subprocess
from heapq import nlargest
from urllib import quote
from datetime import datetime, timedelta
from itertools import islice, count
from collections import Counter
try:
	import cPickle as pickle
except ImportError:
	import pickle
# Flask & co
from flask import Flask, Response
from flask import request, render_template, send_from_directory
#from werkzeug.contrib.cache import SimpleCache
# disco-dop
from discodop.tree import Tree
from discodop.treedraw import DrawTree
from discodop import fragments

CORPUS_DIR = "corpus/"
MINFREQ = 2  # filter out fragments which occur just once or twice
MINNODES = 3  # filter out fragments with only three nodes (CFG productions)
TREELIMIT = 10  # max number of trees to draw in search resuluts
FRAGLIMIT = 250  # max amount of search results for fragment extraction
SENTLIMIT = 1000  # max number of sents/brackets in search results
STYLELANG = 'nl'  # language to use when running style(1)

APP = Flask(__name__)

MORPH_TAGS = re.compile(
		r'([_*A-Z0-9]+)(?:\[[^ ]*\][0-9]?)?((?:-[_A-Z0-9]+)?(?:\*[0-9]+)? )')
FUNC_TAGS = re.compile(r'-[_A-Z0-9]+')
GETLEAVES = re.compile(r" ([^ ()]+)(?=[ )])")
GETFRONTIERNTS = re.compile(r"\(([^ ()]+) \)")
READGRADERE = re.compile(r'([- A-Za-z]+): ([0-9]+(?:\.[0-9]+)?)[\n /]')
AVERAGERE = re.compile(
		r'([a-z]+), average length ([0-9]+(?:\.[0-9]+)?) ([A-Za-z]+)')
PERCENTAGE1RE = re.compile(
		r'([A-Za-z][A-Za-z ()]+) ([0-9]+(?:\.[0-9]+)?)% \([0-9]+\)')
PERCENTAGE2RE = re.compile(
		r'([0-9]+(?:\.[0-9]+)?)% \([0-9]+\) ([A-Za-z ()]+)\n')

# abbreviations for Alpino POS tags
ABBRPOS = {
	'PUNCT': 'PUNCT',
	'COMPLEMENTIZER': 'COMP',
	'PROPER_NAME': 'NAME',
	'PREPOSITION': 'PREP',
	'PRONOUN': 'PRON',
	'DETERMINER': 'DET',
	'ADJECTIVE': 'ADJ',
	'ADVERB': 'ADV',
	'HET_NOUN': 'HET',
	'NUMBER': 'NUM',
	'PARTICLE': 'PRT',
	'ARTICLE': 'ART',
	'NOUN': 'NN',
	'VERB': 'VB'}


@APP.route('/')
@APP.route('/counts')
@APP.route('/trees')
@APP.route('/sents')
@APP.route('/brackets')
@APP.route('/fragments')
def main():
	""" Main search form & results page. """
	output = None
	if request.path != '/':
		output = request.path.lstrip('/')
	elif 'output' in request.args:
		output = request.args['output']
	selected = selectedtexts(request.args)
	if output:
		if output not in DISPATCH:
			return 'Invalid argument', 404
		elif request.args.get('export'):
			return export(request.args, output)
		return Response(stream_template('searchresults.html',
				form=request.args, texts=TEXTS, selectedtexts=selected,
				output=output, results=DISPATCH[output](request.args)))
	return render_template('search.html', form=request.args, output='counts',
			texts=TEXTS, selectedtexts=selected)


@APP.route('/style')
def style():
	""" Use style(1) program to get staticstics for each text. """
	def generate(doexport):
		""" Generator for results. """
		files = glob.glob(os.path.join(CORPUS_DIR, '*.txt'))
		if files and not doexport:
			yield "NB: formatting errors may distort paragraph counts etc.\n\n"
		else:
			files = glob.glob(os.path.join(CORPUS_DIR, '*.t2c.gz'))
			if not doexport:
				yield ("No .txt files found in corpus/\n"
						"Using sentences extracted from parse trees.\n"
						"Supply text files with original formatting\n"
						"to get meaningful paragraph information.\n\n")
		for n, filename in enumerate(sorted(STYLETABLE)):
			name = os.path.basename(filename)
			if doexport:
				if n == 0:
					yield 'text, %s\n' % ', '.join(
							'"%s"' % key for key in sorted(STYLETABLE[name]))
				yield '"%s", %s\n' % (name, ', '.join('%s' % val
						for _, val in sorted(STYLETABLE[name].items())))
			elif n == 0:
				yield '<a href="style?export">Export to CSV</a>'
		if not doexport:
			yield '</pre>'
		# produce a plot for each field
		for a in STYLETABLE:
			fields = sorted(STYLETABLE[a].keys())
			break
		for field in () if doexport else fields:
			data = {a: STYLETABLE[a].get(field, 0) for a in STYLETABLE}
			total = max(data.values())
			if total > 0:
				yield barplot(data, total, field + ':', barstyle='chart1',
						unit='%' if '%' in field else '')

	if 'export' in request.args:
		resp = Response(generate(True),
				mimetype='text/plain')
		resp.headers['Content-Disposition'] = 'attachment; filename=style.csv'
	else:
		resp = Response(stream_template('searchresults.html',
				form=request.args, texts=TEXTS,
				selectedtexts=selectedtexts(request.args),
				output='style', results=generate(False)))
	resp.headers['Cache-Control'] = 'max-age=604800, public'
	#set Expires one day ahead (according to server time)
	resp.headers['Expires'] = (
		datetime.utcnow() + timedelta(7, 0)).strftime(
		'%a, %d %b %Y %H:%M:%S UTC')
	return resp


def export(form, output):
	""" Export search results to a file for download. """
	# NB: no distinction between trees from different texts
	# (Does consttreeviewer support # comments?)
	if output == 'counts':
		results = counts(form, doexport=True)
		filename = 'counts.csv'
	elif output == 'fragments':
		results = fragmentsinresults(form, doexport=True)
		filename = 'fragments.txt'
	else:
		results = (a + '\n' for a in doqueries(
				form, lines=False, doexport=output))
		filename = '%s.txt' % output
	resp = Response(results, mimetype='text/plain')
	resp.headers['Content-Disposition'] = 'attachment; filename=' + filename
	return resp


@APP.route('/draw')
def draw():
	""" Produce a visualization of a tree on a separate page. """
	if 'tree' in request.args:
		return "<pre>%s</pre>" % DrawTree(request.args['tree']).text(
				unicodelines=True, html=True)
	textno, sentno = int(request.args['text']), int(request.args['sent'])
	nofunc = 'nofunc' in request.args
	nomorph = 'nomorph' in request.args
	filename = os.path.join(CORPUS_DIR, TEXTS[textno].replace('.t2c.gz', ''))
	treestr = next(islice(open(filename), sentno - 1, sentno)).decode('utf8')
	return '<pre id="t%s">%s</pre>' % (sentno, DrawTree(
			filterlabels(treestr, nofunc, nomorph)).text(
				unicodelines=True, html=True))


@APP.route('/browse')
def browse():
	""" Browse through trees in a file. """
	chunk = 20  # number of trees to fetch for one request
	if 'text' in request.args and 'sent' in request.args:
		textno = int(request.args['text'])
		sentno = int(request.args['sent']) - 1
		filename = os.path.join(CORPUS_DIR,
				TEXTS[textno].replace('.t2c.gz', ''))
		start = max(0, sentno - sentno % chunk)
		maxtree = min(start + chunk, NUMSENTS[textno])
		nofunc = 'nofunc' in request.args
		nomorph = 'nomorph' in request.args
		lines = islice(open(filename), start, maxtree)
		results = ['<pre id="t%s"%s>%s</pre>' % (n + 1,
				' style="display: none; "' if 'ajax' in request.args else '',
				DrawTree(filterlabels(
					line.decode('utf8'), nofunc, nomorph)).text(
					unicodelines=True, html=True))
				for n, line in enumerate(lines, start)]
		if 'ajax' in request.args:
			return '\n'.join(results)

		prevlink = '<a id=prev>prev</a>'
		if sentno > chunk:
			prevlink = '<a href="browse?text=%d&sent=%d" id=prev>prev</a>' % (
					textno, sentno - chunk + 1)
		nextlink = '<a id=next>next</a>'
		if sentno < NUMSENTS[textno] - chunk:
			nextlink = '<a href="browse?text=%d&sent=%d" id=next>next</a>' % (
					textno, sentno + chunk + 1)
		return render_template('browse.html', textno=textno, sentno=sentno + 1,
				text=TEXTS[textno], totalsents=NUMSENTS[textno], trees=results,
				prevlink=prevlink, nextlink=nextlink, chunk=chunk,
				nofunc=nofunc, nomorph=nomorph,
				mintree=start + 1, maxtree=maxtree)
	return '<ol>\n%s</ol>\n' % '\n'.join(
			'<li><a href="browse?text=%d&sent=1&nomorph">%s</a> '
			'(%d sentences)' % (n, text, NUMSENTS[n])
			for n, text in enumerate(TEXTS))


@APP.route('/favicon.ico')
def favicon():
	""" Serve the favicon. """
	return send_from_directory(os.path.join(APP.root_path, 'static'),
			'treesearch.ico', mimetype='image/vnd.microsoft.icon')


def counts(form, doexport=False):
	""" Produce counts of matches for each text. """
	cnts = Counter()
	relfreq = {}
	sumtotal = 0
	norm = form.get('norm', 'sents')
	gotresult = False
	for n, (textno, results, stderr) in enumerate(doqueries(form, lines=False)):
		if n == 0:
			if doexport:
				yield '"text","count","relfreq"\r\n'
			else:
				url = 'counts?query=%s&norm=%s&texts=%s&export=1' % (
						form['query'], form['norm'], form['texts'])
				yield ('Query: %s\n'
						'NB: the counts reflect the total number of '
						'times a pattern matches for each tree.\n\n'
						'Counts (<a href="%s">export to CSV</a>):\n' % (
						stderr, url))
		cnt = results.count('\n')
		if cnt:
			gotresult = True
		text = TEXTS[textno].replace('.t2c.gz', '')
		cnts[text] = cnt
		total = totalsent = NUMSENTS[textno]
		if norm == 'consts':
			total = NUMCONST[textno]
		elif norm == 'words':
			total = NUMWORDS[textno]
		elif norm != 'sents':
			raise ValueError
		relfreq[text] = 100.0 * cnt / total
		sumtotal += total
		if doexport:
			yield '"%s",%d,%g\r\n' % (text, cnt, relfreq[text])
		else:
			line = "%s%6d    %5.2f %%" % (
					text.ljust(40)[:40], cnt, relfreq[text])
			indices = {int(line[:line.index(':::')])
					for line in results.splitlines()}
			plot = concplot(indices, totalsent)
			if cnt:
				yield line + plot + '\n'
			else:
				yield '<span style="color: gray; ">%s%s</span>\n' % (line, plot)
	if gotresult and not doexport:
		yield ("%s%6d    %5.2f %%\n</span>\n" % (
				"TOTAL".ljust(40),
				sum(cnts.values()),
				100.0 * sum(cnts.values()) / sumtotal))
		yield barplot(relfreq, max(relfreq.values()),
				'Relative frequency of pattern: (count / num_%s * 100)' % norm,
				barstyle='chart1', unit='%')
		#yield barplot(cnts, max(cnts.values()), 'Absolute counts of pattern',
		#		barstyle='chart2')


def trees(form):
	""" Return visualization of parse trees in search results. """
	# TODO: show context of x sentences around result, offer pagination.
	gotresults = False
	for n, (_textno, results, stderr) in enumerate(
			doqueries(form, lines=True)):
		if n == 0:
			# NB: we do not hide function or morphology tags when exporting
			url = 'trees?query=%s&texts=%s&export=1' % (
					form['query'], form['texts'])
			yield ('Query: %s\n'
					'Trees (showing up to %d per text; '
					'export: <a href="%s">plain</a>, '
					'<a href="%s">with line numbers</a>):\n' % (
						stderr, TREELIMIT, url, url + '&linenos=1'))
		for m, line in enumerate(islice(results, TREELIMIT)):
			lineno, text, treestr, match = line.split(":::")
			if m == 0:
				gotresults = True
				yield ("==&gt; %s: [<a href=\"javascript: toggle('n%d'); \">"
						"toggle</a>]\n<span id=n%d>" % (text, n + 1, n + 1))
			cnt = count()
			treestr = treestr.replace(" )", " -NONE-)")
			match = match.strip()
			if match.startswith('('):
				treestr = treestr.replace(match, '%s_HIGH %s' % tuple(
						match.split(None, 1)))
			else:
				match = ' %s)' % match
				treestr = treestr.replace(match, '_HIGH%s' % match)
			tree = Tree.parse(treestr, parse_leaf=lambda _: next(cnt))
			sent = re.findall(r" +([^ ()]+)(?=[ )])", treestr)
			high = list(tree.subtrees(lambda n: n.label.endswith("_HIGH")))
			if high:
				high = high.pop()
				high.label = high.label.rsplit("_", 1)[0]
				high = list(high.subtrees()) + high.leaves()
			try:
				treerepr = DrawTree(tree, sent, highlight=high).text(
						unicodelines=True, html=True)
			except ValueError as err:
				line = "#%s \nERROR: %s\n%s\n%s\n" % (lineno, err, treestr, tree)
			else:
				line = "#%s\n%s\n" % (lineno, treerepr)
			yield line
		yield "</span>"
	if not gotresults:
		yield "No matches."


def sents(form, dobrackets=False):
	""" Return search results as terminals or in bracket notation. """
	gotresults = False
	for n, (textno, results, stderr) in enumerate(
			doqueries(form, lines=True)):
		if n == 0:
			url = '%s?query=%s&texts=%s&export=1' % (
					'trees' if dobrackets else 'sents',
					form['query'], form['texts'])
			yield ('Query: %s\n'
					'Sentences (showing up to %d per text; '
					'export: <a href="%s">plain</a>, '
					'<a href="%s">with line numbers</a>):\n' % (
						stderr, SENTLIMIT, url, url + '&linenos=1'))
		for m, line in enumerate(islice(results, SENTLIMIT)):
			lineno, text, treestr, match = line.rstrip().split(":::")
			if m == 0:
				gotresults = True
				yield ("\n%s: [<a href=\"javascript: toggle('n%d'); \">"
						"toggle</a>] <ol id=n%d>" % (text, n, n))
			treestr = cgi.escape(treestr.replace(" )", " -NONE-)"))
			link = "<a href='/browse?text=%d&sent=%s%s%s'>draw</a>" % (
					textno, lineno,
					'&nofunc' if 'nofunc' in form else '',
					'&nomorph' if 'nomorph' in form else '')
			if dobrackets:
				out = treestr.replace(match,
						"<span class=r>%s</span>" % match)
			else:
				pre, highlight, post = treestr.partition(match)
				out = "%s <span class=r>%s</span> %s " % (
						' '.join(GETLEAVES.findall(pre)),
						' '.join(GETLEAVES.findall(highlight))
								if '(' in highlight else highlight,
						' '.join(GETLEAVES.findall(post)))
			yield "<li>#%s [%s] %s" % (lineno.rjust(6), link, out)
		yield "</ol>"
	if not gotresults:
		yield "No matches."


def brackets(form):
	""" Wrapper. """
	return sents(form, dobrackets=True)


def fragmentsinresults(form, doexport=False):
	""" Extract recurring fragments from search results. """
	gotresults = False
	uniquetrees = set()
	for n, (_, results, stderr) in enumerate(
			doqueries(form, lines=True)):
		if n == 0 and not doexport:
			url = 'fragments?query=%s&texts=%s&export=1' % (
					form['query'], form['texts'])
			yield ('Query: %s\n'
					'Fragments (showing up to %d fragments '
					'in the first %d search results from selected texts; '
					'<a href="%s">Export</a>):\n'
					% (stderr, FRAGLIMIT, SENTLIMIT, url))
		for m, line in enumerate(islice(results, SENTLIMIT)):
			if m == 0:
				gotresults = True
			_, _, treestr, _ = line.rstrip().split(":::")
			treestr = treestr.replace(" )", " -NONE-)") + '\n'
			uniquetrees.add(treestr.encode('utf8'))
	if not gotresults and not doexport:
		yield "No matches."
		return
	# TODO: get counts from whole text (preload)
	with tempfile.NamedTemporaryFile(delete=True) as tmp:
		tmp.writelines(uniquetrees)
		tmp.flush()
		results, approxcounts = fragments.regular((tmp.name, ), 1, 0, 'utf8')
	results = nlargest(FRAGLIMIT, zip(results, approxcounts),
			key=lambda ff: (2 * ff[0].count(')') - ff[0].count(' (')
				- ff[0].count(' )')) * ff[1])
	results = [(frag, freq) for frag, freq in results
			if (2 * frag.count(')')
				- frag.count(' (')
				- frag.count(' )')) > MINNODES and freq > MINFREQ]
	gotresults = False
	if not doexport:
		yield "<ol>"
	for treestr, freq in results:
		gotresults = True
		if doexport:
			yield '%s\t%s\n' % (treestr, freq)
		else:
			link = "<a href='/draw?tree=%s'>draw</a>" % (
					quote(treestr.encode('utf8')))
			treestr = GETLEAVES.sub(' <font color=red>\\1</font>',
					cgi.escape(treestr))
			treestr = GETFRONTIERNTS.sub('(<font color=blue>\\1</font> )',
					treestr)
			yield "<li>freq=%3d [%s] %s" % (freq, link, treestr)
	if not doexport:
		yield "</ol>"
		if not gotresults:
			yield "No fragments with freq > %d & nodes > %d." % (
					MINNODES, MINFREQ)
			return


def doqueries(form, lines=False, doexport=None):
	""" Run tgrep2 on each text """
	selected = selectedtexts(form)
	for n, text in enumerate(TEXTS):
		if n not in selected:
			continue
		if doexport == 'sents':
			fmt = r'%f:%s|%tw\n' if form.get('linenos') else r'%tw\n'
		elif doexport == 'trees' or doexport == 'brackets':
			# NB: no distinction between trees from different texts
			# (Does consttreeviewer support # comments?)
			fmt = r"%f:%s|%w\n" if form.get('linenos') else r"%w\n"
		elif not doexport:
			fmt = r'%s:::%f:::%w:::%h\n'
		else:
			raise ValueError
		cmd = [which('tgrep2'), '-z', '-a',
				'-m', fmt,
				'-c', os.path.join(CORPUS_DIR, text + '.t2c.gz'),
				'static/macros.txt',
				form['query']]
		proc = subprocess.Popen(args=cmd,
				bufsize=-1, shell=False,
				stdout=subprocess.PIPE,
				stderr=subprocess.PIPE)
		out, err = proc.communicate()
		out, err = out.decode('utf8'), err.decode('utf8')
		proc.stdout.close()  # pylint: disable=E1101
		proc.stderr.close()  # pylint: disable=E1101
		proc.wait()  # pylint: disable=E1101
		if lines:
			yield n, filterlabels(out, 'nofunc' in form,
					'nomorph' in form).splitlines(), err
		elif doexport:
			if form.get('linenos'):
				yield re.sub(r'(^|\n)corpus/', '\\1', out)
			else:
				yield out
		else:
			yield n, out, err


def filterlabels(line, nofunc, nomorph):
	""" Optionally remove morphological and grammatical function labels
	from parse tree. """
	if nofunc:
		line = FUNC_TAGS.sub('', line)
	if nomorph:
		line = MORPH_TAGS.sub(lambda g: '%s%s' % (
				ABBRPOS.get(g.group(1), g.group(1)), g.group(2)), line)
	return line


def barplot(data, total, title, barstyle='chart1', width=800.0, unit=''):
	""" A HTML bar plot given a dictionary and max value. """
	result = ['</pre><div class=chart>',
			('<text style="font-family: sans-serif; font-size: 16px; ">'
			'%s</text>' % title)]
	for key in sorted(data, key=data.get, reverse=True):
		result.append('<div class=%s style="width: %gpx" >%s: %g %s</div>' % (
				barstyle if data[key] else 'chart',
				width * data[key] / total, key, data[key], unit))
	result.append('</div>\n<pre>')
	return '\n'.join(result)


def concplot(indices, total, width=800.0):
	""" Draw a concordance plot from a sequence of indices and the total number
	of items. """
	result = ('\t<svg version="1.1" xmlns="http://www.w3.org/2000/svg"'
			' width="%dpx" height="10px" >\n'
			'<rect x=0 y=0 width="%dpx" height=10 '
			'fill=white stroke=black />\n' % (width, width))
	if indices:
		result += ('<g transform="scale(%g, 1)">\n'
				'<path stroke=black d="%s" /></g>') % (width / total,
				''.join('M %d 0v 10' % idx for idx in sorted(indices)))
	return result + '</svg>'


def selectedtexts(form):
	""" Find available texts and parse selected texts argument. """
	selected = set()
	if 'texts' in form:
		for a in filter(None, form['texts'].replace('.', ',').split(',')):
			if '-' in a:
				b, c = a.split('-')
				selected.update(n for n in range(int(b), int(c)))
			else:
				selected.add(int(a))
	else:
		selected.update(range(len(TEXTS)))
	return selected


def preparecorpus():
	""" Produce indexed versions of parse trees in .mrg files """
	files = sorted(glob.glob(os.path.join(CORPUS_DIR, '*.mrg')))
	assert files, 'Expected one or more .mrg files with parse trees in corpus/'
	for a in files:
		if not os.path.exists(a + '.t2c.gz'):
			subprocess.check_call(
					args=[which('tgrep2'), '-p', a, a + '.t2c.gz'],
					shell=False,
					stdout=subprocess.PIPE,
					stderr=subprocess.PIPE)


def getstyletable():
	""" Run style(1) on all files and store results in a dictionary. """
	files = glob.glob(os.path.join(CORPUS_DIR, '*.txt'))
	if not files:
		files = glob.glob(os.path.join(CORPUS_DIR, '*.t2c.gz'))
	styletable = {}
	for filename in sorted(files):
		if filename.endswith('.t2c.gz'):
			tgrep = subprocess.Popen(
					args=[which('tgrep2'), '-t', '-c', filename, '*'],
					shell=False, bufsize=-1, stdout=subprocess.PIPE)
			cmd = [which('style'), '--language', STYLELANG]
			stdin = tgrep.stdout  # pylint: disable=E1101
		else:
			cmd = [which('style'), '--language', STYLELANG, filename]
			stdin = None
		proc = subprocess.Popen(args=cmd, shell=False, bufsize=-1,
				stdin=stdin, stdout=subprocess.PIPE,
				stderr=subprocess.STDOUT)
		if filename.endswith('.t2c.gz'):
			tgrep.wait()  # pylint: disable=E1101
		out = proc.stdout.read()  # pylint: disable=E1101
		if proc.stdin:  # pylint: disable=E1101
			proc.stdin.close()  # pylint: disable=E1101
		proc.stdout.close()  # pylint: disable=E1101
		proc.wait()  # pylint: disable=E1101
		if filename.endswith('.t2c.gz'):
			tgrep.stdout.close()  # pylint: disable=E1101
			tgrep.wait()  # pylint: disable=E1101
		name = os.path.basename(filename)
		styletable[name] = parsestyleoutput(out)
	return styletable


def parsestyleoutput(out):
	""" Extract readability grades, averages, and percentages from style output
	(i.e., all figures except for absolute counts). """
	result = {}
	for key, val in READGRADERE.findall(out):
		result[key.strip()] = float(val)
	for key1, val, key2 in AVERAGERE.findall(out):
		result['average %s per %s' % (key2, key1[:-1])] = float(val)
	m = re.search(r'([0-9]+(?:\.[0-9]+)?) syllables', out)
	if m:
		result['average syllables per word'] = float(m.group(1))
	for key, val in PERCENTAGE1RE.findall(out):
		result['%% %s' % key.strip()] = float(val)
	for val, key in PERCENTAGE2RE.findall(out):
		result['%% %s' % key.strip()] = float(val)
	return result


def getcorpus():
	""" Get list of files and number of lines in them. """
	texts = []
	if os.path.exists('/tmp/treesearchcorpus.pickle'):
		texts, numsents, numconst, numwords, styletable = pickle.load(
				open('/tmp/treesearchcorpus.pickle'))
	files = sorted(glob.glob(os.path.join(CORPUS_DIR, '*.mrg')))
	if set(texts) != {os.path.basename(file) for file in files}:
		texts = [os.path.basename(a) for a in files]
		numsents = [len(open(filename).readlines()) for filename in files]
		numconst = [open(filename).read().count('(') for filename in files]
		numwords = [len(GETLEAVES.findall(open(filename).read()))
				for filename in files]
		styletable = getstyletable()
	pickle.dump((texts, numsents, numconst, numwords, styletable),
			open('/tmp/treesearchcorpus.pickle', 'wb'), protocol=-1)
	return texts, numsents, numconst, numwords, styletable


def stream_template(template_name, **context):
	""" From Flask documentation: pass an iterator to a template. """
	APP.update_template_context(context)
	templ = APP.jinja_env.get_template(template_name)
	result = templ.stream(context)
	result.enable_buffering(5)
	return result


def which(program):
	""" Return first match for program in search path. """
	for path in os.environ.get('PATH', os.defpath).split(":"):
		if path and os.path.exists(os.path.join(path, program)):
			return os.path.join(path, program)
	raise ValueError('%r not found in path; please install it.' % program)


TEXTS, NUMSENTS, NUMCONST, NUMWORDS, STYLETABLE = getcorpus()
fragments.PARAMS.update(disc=False, debug=False, cover=False, complete=False,
		quadratic=False, complement=False, quiet=True, nofreq=False,
		approx=True, indices=False)


# this is redundant but used to support both javascript-enabled /foo
# as well as non-javascript fallback /?output=foo
DISPATCH = {
	'counts': counts,
	'trees': trees,
	'sents': sents,
	'brackets': brackets,
	'fragments': fragmentsinresults,
}


if __name__ == '__main__':
	logging.basicConfig()
	for log in (logging.getLogger(), APP.logger):
		log.setLevel(logging.DEBUG)
		log.handlers[0].setFormatter(logging.Formatter(
				fmt='%(asctime)s %(message)s', datefmt='%Y-%m-%d %H:%M:%S'))
	preparecorpus()
	APP.run(debug=True, host='0.0.0.0')



""" Discontinuous Data-Oriented Parsing (disco-dop).
Main components:

- A parser for Probalistic Linear Context-Free Rewriting Systems (LCFRS),
  as well as Probabilistic Context-Free Grammars (PCFG).
- Facilities to extract and parse with data-oriented parsing (DOP) grammars.
"""



""" Discontinuous tree drawing.

Interesting reference (not used for this code):
T. Eschbach et al., Orth. Hypergraph Drawing, Journal of
Graph Algorithms and Applications, 10(2) 141--157 (2006)149.
http://jgaa.info/accepted/2006/EschbachGuentherBecker2006.10.2.pdf
"""

from __future__ import division, print_function, unicode_literals
import sys
import codecs
from cgi import escape
from collections import defaultdict, OrderedDict
from operator import itemgetter
from itertools import count
from discodop.tree import Tree
if sys.version[0] >= '3':
	basestring = str  # pylint: disable=W0622,C0103
	from builtins import zip as izip  # pylint: disable=F0401
else:
	from itertools import izip


USAGE = """Usage: %s [<treebank>...] [options]
Options:
  --fmt=x          Specify corpus format. Options: export, bracket,
                   discbracket, alpino.
  --encoding=enc   Specify a different encoding than the default UTF-8.
  --functions=x    'leave'=default: leave syntactic labels as is,
                   'remove': strip functions off labels,
                   'add': evaluate both syntactic categories and functions,
                   'replace': only evaluate grammatical functions.
  --morphology=x   'no'=default: only evaluate POS tags,
                   'add': concatenate morphology tags to POS tags,
                   'replace': replace POS tags with morphology tags,
                   'between': add morphological node between POS tag and word.
  --abbr           abbreviate labels longer than 5 characters.
  --plain          disable ANSI colors.
If no treebank is given, input is read from standard input; format is detected.
If more than one treebank is specified, trees will be displayed in parallel.
Pipe the output through 'less -R' to preserve the colors. """ % sys.argv[0]
ANSICOLOR = {
		'black': 30,
		'red': 31,
		'green': 32,
		'yellow': 33,
		'blue': 34,
		'magenta': 35,
		'cyan': 36,
		'white': 37,
}


class DrawTree(object):
	""" Visualize a discontinuous tree in various formats.

	``DrawTree(tree, sent=None, highlight=(), abbr=False)``
	creates an object from which different visualizations can be created.

	:param tree: a Tree object or a string.
	:param sent: a list of words (strings). If sent is not given, and the tree
		contains non-integers as leaves, a continuous phrase-structure tree
		is assumed. If sent is not given and the tree contains only indices
		as leaves, the indices are displayed as placeholder terminals.
	:param highlight: Optionally, a sequence of Tree objects in `tree` which
		should be highlighted. Has the effect of only applying colors to nodes
		in this sequence (nodes should be given as Tree objects, terminals as
		indices).
	:param abbr: when True, abbreviate labels longer than 5 characters.

	>>> print(DrawTree('(S (NP Mary) (VP walks))'))
		  S
	  ____|____
	 NP        VP
	 |         |
	Mary     walks
	"""
	def __init__(self, tree, sent=None, highlight=(), abbr=False):
		self.tree = tree
		self.sent = sent
		if isinstance(tree, basestring):
			self.tree = Tree(tree)
		if sent is None:
			leaves = self.tree.leaves()
			if (leaves and not any(len(a) == 0 for a in self.tree.subtrees())
					and all(isinstance(a, int) for a in leaves)):
				self.sent = [str(a) for a in leaves]
			else:
				# this deals with empty nodes (frontier non-terminals)
				# and multiple/mixed terminals under non-terminals.
				self.tree = self.tree.copy(True)
				self.sent = []
				for a in self.tree.subtrees():
					if len(a) == 0:
						a.append(len(self.sent))
						self.sent.append(None)
					elif any(not isinstance(b, Tree) for b in a):
						for n, b in enumerate(a):
							if not isinstance(b, Tree):
								a[n] = len(self.sent)
								self.sent.append('%s' % b)
		if abbr:
			if self.tree is tree:
				self.tree = self.tree.copy(True)
			for n in self.tree.subtrees(lambda x: len(x.label) > 5):
				n.label = n.label[:4] + u'\u2026'  # unicode '...' ellipsis
		self.highlight = set()
		self.nodes, self.coords, self.edges = self.nodecoords(
				self.tree, self.sent, highlight)

	def __str__(self):
		return self.text()

	def __repr__(self):
		return '\n'.join('%d: coord=%r, parent=%r, node=%s' % (
						n, self.coords[n], self.edges.get(n), self.nodes[n])
					for n in sorted(self.nodes))

	def nodecoords(self, tree, sent, highlight):
		""" Resolve placement of nodes for drawing discontinuous trees
		programmatically. Objective:

		- Produce coordinates for a non-overlapping placement of nodes and
			horizontal lines.
		- Order edges so that crossing edges cross a minimal number of previous
			horizontal lines (never vertical lines).

		Approach:

		- bottom up level order traversal (start at terminals)
		- at each level, identify nodes which cannot be on the same row
		- identify nodes which cannot be in the same column
		- place nodes into a grid at (row, column)
		- order child-parent edges with crossing edges last

		Coordinates are (row, column); the origin (0, 0) is at the top left;
		the root node is on row 0. Coordinates do not consider the size of a
		node (which depends on font, &c), so the width of a column of the grid
		should be automatically determined by the element with the greatest
		width in that column. Alternatively, the integer coordinates could be
		converted to coordinates in which the distances between adjacent nodes
		are non-uniform.

		Produces tuple (nodes, coords, edges) where:

		- nodes[id]: Tree object for the node with this integer id
		- coords[id]: (n, m) coordinate where to draw node with id in the grid
		- edges[id]: parent id of node with this id (ordered dictionary)
		"""
		def findcell(m, matrix, startoflevel, children):
			""" find vacant row, column index
			iterate over current rows for this level (try lowest first)
			and look for cell between first and last child of this node,
			add new row to level if no free row available. """
			candidates = [a for _, a in children[m]]
			center = min(candidates) + (max(candidates) - min(candidates)) // 2
			initi = initj = i = j = center

			for idx in range(startoflevel, len(matrix) + 1):
				if idx == len(matrix):
					# need to add a new row
					matrix.append([vertline if a not in (corner, None)
							else None for a in matrix[-1]])
				row = matrix[idx]
				i, j = initi, initj
				# place unaries directly above child; no restrictions for root
				if len(children[m]) == 1 or ids[m] == 0:
					return idx, i
				elif all(a in (None, vertline) for a in
						row[min(children[m], key=itemgetter(1))[1]:
								max(children[m], key=itemgetter(1))[1] + 1]):
					# find free column
					while zeroindex < j or i < scale * len(sent):
						if scale * len(sent) > i and matrix[idx][i] is None:
							return idx, i
						elif zeroindex < j and matrix[idx][j] is None:
							return idx, j
						i += 1
						j -= 1
			raise ValueError('could not find a free cell.')

		leaves = tree.leaves()
		assert all(isinstance(n, int) for n in leaves), (
				'All leaves must be integer indices.')
		assert len(leaves) == len(set(leaves)), (
				'Indices must occur at most once.')
		assert all(0 <= n < len(sent) for n in leaves), (
					'All leaves must be in the interval 0..n with n=len(sent)'
					'\nlength: %d leaves: %r\nsent: %s' % (
					len(sent), tree.leaves(), sent))
		vertline, corner = -1, -2  # constants
		tree = tree.copy(True)
		for a in tree.subtrees(lambda n: n and isinstance(n[0], Tree)):
			a.sort(key=lambda n: min(n.leaves()))
		scale = 10
		crossed = set()
		zeroindex = min(tree.leaves())
		# internal nodes and lexical nodes (no frontiers)
		positions = tree.treepositions()
		depth = max(map(len, positions)) + 1
		childcols = defaultdict(set)
		matrix = [[None] * (len(sent) * scale)]
		nodes = {}
		ids = {a: n for n, a in enumerate(positions)}
		self.highlight.update(n for a, n in ids.items()
				if not highlight or tree[a] in highlight)
		levels = {n: [] for n in range(depth)}
		preterminals = []
		terminals = []
		for a in positions:
			node = tree[a]
			if isinstance(node, Tree):
				if (node and isinstance(node[0], int)
						and sent[node[0]] is not None):
					preterminals.append(a)
				else:
					levels[len(a)].append(a)
			else:
				terminals.append(a)

		for n in levels:
			levels[n].sort(key=lambda n: (max(tree[n].leaves())
					- min(tree[n].leaves()), len(n)), reverse=True)
		preterminals.sort()
		levels[max(levels) + 1] = preterminals
		terminals.sort()
		positions = set(positions)

		for m in terminals:
			i = int(tree[m]) * scale
			assert matrix[0][i] is None, (
					matrix[0][i], m, i)
			matrix[0][i] = ids[m]
			nodes[ids[m]] = sent[tree[m]]
			if nodes[ids[m]] is None:
				nodes[ids[m]] = '...'
			positions.remove(m)
			childcols[m[:-1]].add((0, i))

		# add other nodes centered on their children,
		# if the center is already taken, back off
		# to the left and right alternately, until an empty cell is found.
		for n in sorted(levels, reverse=True):
			nodesatdepth = levels[n]
			startoflevel = len(matrix)
			matrix.append([vertline if a not in (corner, None) else None
					for a in matrix[-1]])
			for m in nodesatdepth:  # [::-1]:
				if n < depth - 1 and childcols[m]:
					_, pivot = min(childcols[m], key=itemgetter(1))
					if ({a[:-1] for row in matrix[:-1] for a in row[:pivot]
							if isinstance(a, tuple)} &
						{a[:-1] for row in matrix[:-1] for a in row[pivot:]
							if isinstance(a, tuple)}):
						crossed.add(m)

				rowindex, i = findcell(m, matrix, startoflevel, childcols)
				positions.remove(m)

				# block positions where children of this node branch out
				for _, x in childcols[m]:
					matrix[rowindex][x] = corner
				# node itself
				#FIXME: assert matrix[rowindex][i] in (None, corner), (
				#		matrix[rowindex][i], m, str(tree), ' '.join(sent))
				matrix[rowindex][i] = ids[m]
				nodes[ids[m]] = tree[m]
				# add column to the set of children for its parent
				if m != ():
					childcols[m[:-1]].add((rowindex, i))
			#print('level', n)
			#for row in matrix: print(''.join(('%4s' % a)[-2:] for a in row))
		assert len(positions) == 0

		# remove unused columns, right to left
		for m in range(scale * len(sent) - 1, -1, -1):
			if not any(isinstance(row[m], (Tree, int))
					for row in matrix):
				for row in matrix:
					del row[m]

		# remove unused rows
		matrix = [row for row in matrix
				if not all(a is None for a in row)][::-1]

		# collect coordinates of nodes
		coords = {}
		for n, _ in enumerate(matrix):
			for m, i in enumerate(matrix[n]):
				if isinstance(i, int) and i >= 0:
					coords[i] = n, m

		# move crossed edges last
		positions = sorted([a for level in levels.values()
				for a in level] + preterminals,
				key=lambda a: a[:-1] in crossed)

		# collect edges from node to node
		edges = OrderedDict()
		for i in reversed(positions):
			for j, _ in enumerate(tree[i]):
				edges[ids[i + (j, )]] = ids[i]
		#for row in matrix: print(', '.join('%4s' % a for a in row))
		return nodes, coords, edges

	def svg(self, nodecolor='blue', leafcolor='red'):
		""" :returns: SVG representation of a discontinuous tree. """
		fontsize = 12
		hscale = 40
		vscale = 25
		hstart = vstart = 20
		width = max(col for _, col in self.coords.values())
		height = max(row for row, _ in self.coords.values())
		result = ['<svg version="1.1" xmlns="http://www.w3.org/2000/svg" '
				'width="%dem" height="%dem" viewBox="%d %d %d %d">' % (
						width * 3,
						height * 2.5,
						-hstart, -vstart,
						width * hscale + 3 * hstart,
						height * vscale + 3 * vstart)]

		children = defaultdict(set)
		for n in self.nodes:
			if n:
				children[self.edges[n]].add(n)

		# horizontal branches from nodes to children
		for node in self.nodes:
			if not children[node]:
				continue
			y, x = self.coords[node]
			x *= hscale
			y *= vscale
			x += hstart
			y += vstart + fontsize // 2
			childx = [self.coords[c][1] for c in children[node]]
			xmin = hstart + hscale * min(childx)
			xmax = hstart + hscale * max(childx)
			result.append(
				'\t<polyline style="stroke:black; stroke-width:1; fill:none;" '
					'points="%g,%g %g,%g" />' % (
					xmin, y, xmax, y))
			result.append(
				'\t<polyline style="stroke:black; stroke-width:1; fill:none;" '
					'points="%g,%g %g,%g" />' % (
					x, y, x, y - fontsize // 3))

		# vertical branches from children to parents
		for child, parent in self.edges.items():
			y, _ = self.coords[parent]
			y *= vscale
			y += vstart + fontsize // 2
			childy, childx = self.coords[child]
			childx *= hscale
			childy *= vscale
			childx += hstart
			childy += vstart - fontsize
			result.append(
				'\t<polyline style="stroke:white; stroke-width:10; fill:none;"'
				' points="%g,%g %g,%g" />' % (childx, childy, childx, y + 5))
			result.append(
				'\t<polyline style="stroke:black; stroke-width:1; fill:none;" '
				'points="%g,%g %g,%g" />' % (childx, childy, childx, y))

		# write nodes with coordinates
		for n, (row, column) in self.coords.items():
			node = self.nodes[n]
			x = column * hscale + hstart
			y = row * vscale + vstart
			if n in self.highlight:
				color = nodecolor if isinstance(node, Tree) else leafcolor
			else:
				color = 'black'
			result.append('\t<text style="text-anchor: middle; fill: %s; '
					'font-size: %dpx;" x="%g" y="%g">%s</text>' % (
					color, fontsize, x, y,
					escape(node.label if isinstance(node, Tree) else node)))

		result += ['</svg>']
		return '\n'.join(result)

	def text(self, nodedist=1, unicodelines=False, html=False, ansi=False,
				nodecolor='blue', leafcolor='red'):
		""" :returns: ASCII art for a discontinuous tree.

		:param unicodelines: whether to use Unicode line drawing characters
			instead of plain (7-bit) ASCII.
		:param html: whether to wrap output in html code (default plain text).
		:param ansi: whether to produce colors with ANSI escape sequences
			(only effective when html==False).
		:param leafcolor, nodecolor: specify colors of leaves and phrasal
			nodes. """

		if unicodelines:
			horzline = u'\u2500'
			leftcorner = u'\u250c'
			rightcorner = u'\u2510'
			vertline = u' \u2502 '
			tee = horzline + u'\u252C' + horzline
			bottom = horzline + u'\u2534' + horzline
			cross = horzline + u'\u253c' + horzline
		else:
			horzline = '_'
			leftcorner = rightcorner = ' '
			vertline = ' | '
			tee = 3 * horzline
			cross = bottom = '_|_'

		def crosscell(cur, x=vertline):
			""" Overwrite center of this cell with a vertical branch. """
			splitl = len(cur) - len(cur) // 2 - len(x) // 2 - 1
			lst = list(cur)
			lst[splitl:splitl + len(x)] = list(x)
			return ''.join(lst)

		result = []
		matrix = defaultdict(dict)
		maxnode = defaultdict(lambda: 3)
		maxcol = 0
		minchildcol = {}
		maxchildcol = {}
		childcols = defaultdict(set)
		for a in self.nodes:
			row, column = self.coords[a]
			matrix[row][column] = a
			maxcol = max(maxcol, column)
			maxnode[column] = max(maxnode[column], len(self.nodes[a].label
					if isinstance(self.nodes[a], Tree) else self.nodes[a]))
			if a not in self.edges:
				continue  # e.g., root
			parent = self.edges[a]
			childcols[parent].add((row, column))
			minchildcol[parent] = min(minchildcol.get(parent, column), column)
			maxchildcol[parent] = max(maxchildcol.get(parent, column), column)
		# bottom up level order traversal
		for row in sorted(matrix, reverse=True):
			noderow = [''.center(maxnode[col]) for col in range(maxcol + 1)]
			branchrow = [''.center(maxnode[col]) for col in range(maxcol + 1)]
			for col in matrix[row]:
				n = matrix[row][col]
				node = self.nodes[n]
				if isinstance(node, Tree):
					text = node.label
					# horizontal branch towards children for this node
					if n in minchildcol and minchildcol[n] < maxchildcol[n]:
						i, j = minchildcol[n], maxchildcol[n]
						a, b = (maxnode[i] + 1) // 2 - 1, maxnode[j] // 2
						branchrow[i] = ((' ' * a) + leftcorner).ljust(
								maxnode[i], horzline)
						branchrow[j] = (rightcorner + (' ' * b)).rjust(
								maxnode[j], horzline)
						for i in range(minchildcol[n] + 1, maxchildcol[n]):
							if i == col and any(
									a == i for _, a in childcols[n]):
								line = cross
							elif i == col:
								line = bottom
							elif any(a == i for _, a in childcols[n]):
								line = tee
							else:
								line = horzline
							branchrow[i] = line.center(maxnode[i], horzline)
					else:  # if n and n in minchildcol:
						branchrow[col] = crosscell(branchrow[col])
				else:
					text = node
				text = text.center(maxnode[col])
				if html:
					text = escape(text)
					if n in self.highlight:
						text = '<font color=%s>%s</font>' % (
								nodecolor if isinstance(node, Tree)
								else leafcolor, text)
				elif ansi and n in self.highlight:
					text = '\x1b[%d;1m%s\x1b[0m' % (
							ANSICOLOR[nodecolor] if isinstance(node, Tree)
							else ANSICOLOR[leafcolor], text)
				noderow[col] = text
			# for each column, if there is a node below us which has a parent
			# above us, draw a vertical branch in that column.
			if row != max(matrix):
				for n, (childrow, col) in self.coords.items():
					if (n > 0 and
							self.coords[self.edges[n]][0] < row < childrow):
						branchrow[col] = crosscell(branchrow[col])
						if col not in matrix[row]:
							noderow[col] = crosscell(noderow[col])
				branchrow = [a + ((a[-1] if a[-1] != ' ' else b[0]) * nodedist)
						for a, b in zip(branchrow, branchrow[1:] + [' '])]
				result.append(''.join(branchrow))
				result.append((' ' * nodedist).join(noderow))
			else:
				result.append((' ' * nodedist).join(noderow))
		return '\n'.join(reversed(result)) + '\n'

	def tikzmatrix(self, nodecolor='blue', leafcolor='red'):
		""" Produce TiKZ code for use with LaTeX. PDF can be produced with
		pdflatex. Uses TiKZ matrices meaning that nodes are put into a fixed
		grid. Where the cells of each column all have the same width. """
		result = ['%% %s\n%% %s' % (self.tree, ' '.join(self.sent)),
			r'''\begin{tikzpicture}[scale=1, minimum height=1.25em,
			text height=1.25ex, text depth=.25ex,
			inner sep=0mm, node distance=1mm]''',
		r'\footnotesize\sffamily',
		r'\matrix[row sep=0.5cm,column sep=0.1cm] {']

		# write matrix with nodes
		matrix = defaultdict(dict)
		maxcol = 0
		for a in self.nodes:
			row, column = self.coords[a]
			matrix[row][column] = a
			maxcol = max(maxcol, column)

		for row in sorted(matrix):
			line = []
			for col in range(maxcol + 1):
				if col in matrix[row]:
					i = matrix[row][col]
					node = self.nodes[i]
					if isinstance(node, Tree):
						color = nodecolor if i in self.highlight else 'black'
						label = latexlabel(node.label)
					else:
						color = leafcolor if i in self.highlight else 'black'
						label = node
					line.append(r'\node [%s] (n%d) { %s };' % (
							color, i, label))
				line.append('&')
			# new row: skip last column char '&', add newline
			result.append(' '.join(line[:-1]) + r' \\')
		result += ['};']

		shift = -0.5
		# write branches from node to node
		for child, parent in self.edges.items():
			result.append(
				'\\draw [white, -, line width=6pt] (n%d)  +(0, %g) -| (n%d);'
				'	\\draw (n%d) -- +(0, %g) -| (n%d);' % (
				parent, shift, child, parent, shift, child))

		result += [r'\end{tikzpicture}']
		return '\n'.join(result)

	def tikznode(self, nodecolor='blue', leafcolor='red'):
		""" Produce TiKZ code to draw a tree. Nodes are drawn with the
		\\node command so they can have arbitrary coordinates. """
		result = ['%% %s\n%% %s' % (self.tree, ' '.join(self.sent)),
			r'''\begin{tikzpicture}[scale=0.75, minimum height=1.25em,
			text height=1.25ex, text depth=.25ex,
			inner sep=0mm, node distance=1mm]''',
		r'\footnotesize\sffamily',
		r'\path']

		bottom = max(row for row, _ in self.coords.values())
		# write nodes with coordinates
		for n, (row, column) in self.coords.items():
			node = self.nodes[n]
			if n in self.highlight:
				color = nodecolor if isinstance(node, Tree) else leafcolor
			else:
				color = 'black'
			result.append('\t(%d, %d) node [%s] (n%d) {%s}'
					% (column, bottom - row,
					color,
					n, latexlabel(node.label)
					if isinstance(node, Tree) else node))
		result += [';']

		shift = -0.5
		# write branches from node to node
		for child, parent in self.edges.items():
			result.append(
				'\\draw [white, -, line width=6pt] (n%d)  +(0, %g) -| (n%d);'
				'	\\draw (n%d) -- +(0, %g) -| (n%d);' % (
				parent, shift, child, parent, shift, child))

		result += [r'\end{tikzpicture}']
		return '\n'.join(result)


def latexlabel(label):
	""" quote/format label for latex """
	newlabel = label.replace('$', r'\$').replace('[', '(').replace('_', r'\_')
	# turn binarization marker into subscripts in math mode
	if '|' in newlabel:
		cat, siblings = newlabel.split('|', 1)
		siblings = siblings.strip('<>')
		if '^' in siblings:
			siblings, parents = siblings.split('^')
			newlabel = '$ \\textsf{%s}_\\textsf{%s}^\\textsf{%s} $' % (
					cat, siblings[1:-1], parents)
		else:
			newlabel = '$ \\textsf{%s}_\\textsf{%s} $' % (
					cat, siblings)
	else:
		newlabel = newlabel.replace('<', '$<$').replace('>', '$>$')
	return newlabel


def test():
	""" Do some tests. """
	trees = '''(ROOT (S (ADV 0) (VVFIN 1) (NP (PDAT 2) (NN 3)) (PTKNEG 4) \
				(PP (APPRART 5) (NN 6) (NP (ART 7) (ADJA 8) (NN 9)))) ($. 10))
			(S (NP (NN 1) (EX 3)) (VP (VB 0) (JJ 2)))
			(S (VP (PDS 0) (ADV 3) (VVINF 4)) (PIS 2) (VMFIN 1))
			(top (du (comp 0) (smain (noun 1) (verb 2) (inf (verb 8) (inf \
				(adj 3) (pp (prep 4) (np (det 5) (noun 6))) (part 7) (verb 9) \
				(pp (prep 10) (np (det 11) (noun 12) (pp (prep 13) (mwu \
				(noun 14) (noun 15))))))))) (punct 16))
			(top (smain (noun 0) (verb 1) (inf (verb 5) (inf (np (det 2) \
				(adj 3) (noun 4)) (verb 6) (pp (prep 7) (noun 8))))) (punct 9))
			(top (smain (noun 0) (verb 1) (noun 2) (inf (adv 3) (verb 4))) \
				(punct 5))
			(top (punct 5) (du (smain (noun 0) (verb 1) (ppart (np (det 2) \
				(noun 3)) (verb 4))) (conj (sv1 (conj (noun 6) (vg 7) (np \
				(det 8) (noun 9))) (verb 10) (noun 11) (part 12)) (vg 13) \
				(sv1 (verb 14) (ti (comp 19) (inf (np (conj (det 15) (vg 16) \
				(det 17)) (noun 18)) (verb 20)))))) (punct 21))
			(top (punct 10) (punct 16) (punct 18) (smain (np (det 0) (noun 1) \
				(pp (prep 2) (np (det 3) (noun 4)))) (verb 5) (adv 6) (np \
				(noun 7) (noun 8)) (part 9) (np (det 11) (noun 12) (pp \
				(prep 13) (np (det 14) (noun 15)))) (conj (vg 20) (ppres \
				(adj 17) (pp (prep 22) (np (det 23) (adj 24) (noun 25)))) \
				(ppres (adj 19)) (ppres (adj 21)))) (punct 26))
			(top (punct 10) (punct 11) (punct 16) (smain (np (det 0) \
				(noun 1)) (verb 2) (np (det 3) (noun 4)) (adv 5) (du (cp \
				(comp 6) (ssub (noun 7) (verb 8) (inf (verb 9)))) (du \
				(smain (noun 12) (verb 13) (adv 14) (part 15)) (noun 17)))) \
				(punct 18) (punct 19))
			(top (smain (noun 0) (verb 1) (inf (verb 8) (inf (verb 9) (inf \
				(adv 2) (pp (prep 3) (noun 4)) (pp (prep 5) (np (det 6) \
				(noun 7))) (verb 10))))) (punct 11))
			(top (smain (noun 0) (verb 1) (pp (prep 2) (np (det 3) (adj 4) \
				(noun 5) (rel (noun 6) (ssub (noun 7) (verb 10) (ppart \
				(adj 8) (part 9) (verb 11))))))) (punct 12))
			(top (smain (np (det 0) (noun 1)) (verb 2) (ap (adv 3) (num 4) \
				(cp (comp 5) (np (det 6) (adj 7) (noun 8) (rel (noun 9) (ssub \
				(noun 10) (verb 11) (pp (prep 12) (np (det 13) (adj 14) \
				(adj 15) (noun 16))))))))) (punct 17))
			(top (smain (np (det 0) (noun 1)) (verb 2) (adv 3) (pp (prep 4) \
				(np (det 5) (noun 6)) (part 7))) (punct 8))
			(top (punct 7) (conj (smain (noun 0) (verb 1) (np (det 2) \
				(noun 3)) (pp (prep 4) (np (det 5) (noun 6)))) (smain \
				(verb 8) (np (det 9) (num 10) (noun 11)) (part 12)) (vg 13) \
				(smain (verb 14) (noun 15) (pp (prep 16) (np (det 17) \
				(noun 18) (pp (prep 19) (np (det 20) (noun 21))))))) \
				(punct 22))
			(top (smain (np (det 0) (noun 1) (rel (noun 2) (ssub (np (num 3) \
				(noun 4)) (adj 5) (verb 6)))) (verb 7) (ppart (verb 8) (pp \
				(prep 9) (noun 10)))) (punct 11))
			(top (conj (sv1 (np (det 0) (noun 1)) (verb 2) (ppart (verb 3))) \
				(vg 4) (sv1 (verb 5) (pp (prep 6) (np (det 7) (adj 8) \
				(noun 9))))) (punct 10))
			(top (smain (noun 0) (verb 1) (np (det 2) (noun 3)) (inf (adj 4) \
				(verb 5) (cp (comp 6) (ssub (noun 7) (adv 8) (verb 10) (ap \
				(num 9) (cp (comp 11) (np (det 12) (adj 13) (noun 14) (pp \
				(prep 15) (conj (np (det 16) (noun 17)) (vg 18) (np \
				(noun 19))))))))))) (punct 20))
			(top (punct 8) (smain (noun 0) (verb 1) (inf (verb 5) \
				(inf (verb 6) (conj (inf (pp (prep 2) (np (det 3) (noun 4))) \
				(verb 7)) (inf (verb 9)) (vg 10) (inf (verb 11)))))) \
				(punct 12))
			(top (smain (verb 2) (noun 3) (adv 4) (ppart (np (det 0) \
				(noun 1)) (verb 5))) (punct 6))
			(top (conj (smain (np (det 0) (noun 1)) (verb 2) (adj 3) (pp \
				(prep 4) (np (det 5) (noun 6)))) (vg 7) (smain (np (det 8) \
				(noun 9) (pp (prep 10) (np (det 11) (noun 12)))) (verb 13) \
				(pp (prep 14) (np (det 15) (noun 16))))) (punct 17))
			(top (conj (smain (noun 0) (verb 1) (inf (ppart (np (noun 2) \
				(noun 3)) (verb 4)) (verb 5))) (vg 6) (smain (noun 7) \
				(inf (ppart (np (det 8) (noun 9)))))) (punct 10))
			(A (B1 (t 6) (t 13)) (B2 (t 3) (t 7) (t 10))  (B3 (t 1) \
				(t 9) (t 11) (t 14) (t 16)) (B4 (t 0) (t 5) (t 8)))
			(A (B1 6 13) (B2 3 7 10)  (B3 1 \
				9 11 14 16) (B4 0 5 8))
			(VP (VB 0) (PRT 2))
			(VP (VP 0 3) (NP (PRP 1) (NN 2)))
			(ROOT (S (VP_2 (PP (APPR 0) (ART 1) (NN 2) (PP (APPR 3) (ART 4) \
				(ADJA 5) (NN 6))) (ADJD 10) (PP (APPR 11) (NN 12)) (VVPP 13)) \
				(VAFIN 7) (NP (ART 8) (NN 9))) ($. 14))'''
	sents = '''Leider stehen diese Fragen nicht im Vordergrund der \
				augenblicklichen Diskussion .
			is Mary happy there
			das muss man jetzt machen
			Of ze had gewoon met haar vriendinnen rond kunnen slenteren in de \
				buurt van Trafalgar Square .
			Het had een prachtige dag kunnen zijn in Londen .
			Cathy zag hen wild zwaaien .
			Het was een spel geworden , zij en haar vriendinnen kozen iemand \
				uit en probeerden zijn of haar nationaliteit te raden .
			Elk jaar in het hoogseizoen trokken daar massa's toeristen \
				voorbij , hun fototoestel in de aanslag , pratend , gillend \
				en lachend in de vreemdste talen .
			Haar vader stak zijn duim omhoog alsof hij wilde zeggen : " het \
				komt wel goed , joch " .
			Ze hadden languit naast elkaar op de strandstoelen kunnen gaan \
				liggen .
			Het hoorde bij de warme zomerdag die ze ginds achter had gelaten .
			De oprijlaan was niet meer dan een hobbelige zandstrook die zich \
				voortslingerde tussen de hoge grijze boomstammen .
			Haar moeder kleefde bijna tegen het autoraampje aan .
			Ze veegde de tranen uit haar ooghoeken , tilde haar twee koffers \
				op en begaf zich in de richting van het landhuis .
			Het meisje dat vijf keer juist raadde werd getrakteerd op ijs .
			Haar neus werd platgedrukt en leek op een jonge champignon .
			Cathy zag de BMW langzaam verdwijnen tot hij niet meer was dan \
				een zilveren schijnsel tussen de bomen en struiken .
			Ze had met haar moeder kunnen gaan winkelen , zwemmen of \
				terrassen .
			Dat werkwoord had ze zelf uitgevonden .
			De middagzon hing klein tussen de takken en de schaduwen van de \
				wolken drentelden over het gras .
			Zij zou mams rug ingewreven hebben en mam de hare .
			0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16
			0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16
			0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16
			0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16
			Mit einer Messe in der Sixtinischen Kapelle ist das Konklave \
				offiziell zu Ende gegangen .'''
	trees = [Tree.parse(a, parse_leaf=int) for a in trees.splitlines()]
	sents = [a.split() for a in sents.splitlines()]
	sents.extend([['Wake', None, 'up'],
		[None, 'your', 'friend', None]])
	for n, (tree, sent) in enumerate(zip(trees, sents)):
		drawtree = DrawTree(tree, sent)
		print('\ntree, sent', n, tree,
				' '.join('...' if a is None else a for a in sent),
				repr(drawtree),
				sep='\n')
		try:
			print(drawtree.text(unicodelines=True, ansi=True), sep='\n')
		except (UnicodeDecodeError, UnicodeEncodeError):
			print(drawtree.text(unicodelines=False, ansi=False), sep='\n')


def main():
	""" Text-based tree viewer. """
	from discodop.treebank import getreader, incrementaltreereader, splitpath
	from getopt import gnu_getopt, GetoptError
	flags = ('test', 'help', 'abbr', 'plain')
	options = ('fmt=', 'encoding=', 'functions=', 'morphology=')
	try:
		opts, args = gnu_getopt(sys.argv[1:], '', flags + options)
	except GetoptError as err:
		print('error: %s\n%s' % (err, USAGE))
		sys.exit(2)
	opts = dict(opts)
	if '--test' in opts:
		test()
		return
	elif '--help' in opts:
		print(USAGE)
		return
	if args:
		reader = getreader(opts.get('--fmt', 'export'))
		corpora = []
		for path in args:
			corpus = reader(*splitpath(path),
					encoding=opts.get('--encoding', 'utf8'),
					functions=opts.get('--functions'),
					morphology=opts.get('--morphology'))
			corpora.append(zip(corpus.parsed_sents(),
					zip(corpus.parsed_sents().values(),
						corpus.sents().values())))
		numsents = len(corpus.sents())
		print('Viewing:', *args)
		for n, (sentid, (tree, sent)) in enumerate(corpora[0], 1):
			print('%d of %s (sentid=%s; len=%d):' % (
					n, numsents, sentid, len(sent)))
			for trees, sents in corpora:
				tree, sent = trees[sentid], sents[sentid]
				print(DrawTree(tree, sent, abbr='--abbr' in opts
						).text(unicodelines=True, ansi=not '--plain' in opts))
	else:  # read from stdin + detect format
		sys.stdin = codecs.getreader(opts.get('--encoding', 'utf8'))(sys.stdin)
		trees = incrementaltreereader(sys.stdin,
				morphology=opts.get('--morphology'),
				functions=opts.get('--functions'))
		corpora = [izip(count(1), trees)]
		try:
			for n, (sentid, (tree, sent)) in enumerate(corpora[0], 1):
				print('%d (sentid=%s; len=%d):' % (
						n, sentid, len(sent)))
				print(DrawTree(tree, sent, abbr='--abbr' in opts
						).text(unicodelines=True, ansi=not '--plain' in opts))
		except (IOError, KeyboardInterrupt):
			pass


if __name__ == '__main__':
	main()



# -*- coding: UTF-8 -*-
""" Functions related to dealing with unknown words
and smoothing of lexical probabilities.

Rare words in the training set are replaced with word signatures, such that
unknown words can receive similar tags. Given a function to produce such
signatures from words, the flow is as follows:

Simple lexical smoothing:
	getunknownwordmodel (get statistics)
	replaceraretrainwords (adjust trees)
	[ read off grammar ]
	simplesmoothlexicon (add extra lexical productions)

Non-simple (untested):
	getunknownwordmodel
	getlexmodel
	replaceraretrainwords
	[ read off grammar ]
	smoothlexicon

During parsing:
	replaceraretestwords (only give known words and signatures to parser)
"""
from __future__ import division, print_function, unicode_literals
import re
from collections import defaultdict, Counter as multiset
from fractions import Fraction

UNK = '_UNK'


def getunknownwordmodel(tagged_sents, unknownword,
		unknownthreshold, openclassthreshold):
	""" Compute an unknown word model that smooths lexical probabilities
	for unknown & rare words.

	:param tagged_sents: the sentences from the training set with the gold POS
			tags from the treebank.
	:param unknownword: a function that returns a signature for a given word;
			e.g., "eschewed" => "_UNK-L-d".
	:param unknownthreshold: words with frequency lower than or equal to this
			are replaced by their signature.
	:param openclassthreshold: tags that rewrite to at least this much word
			types are considered to be open class categories. """
	wordsfortag = defaultdict(set)
	tags = multiset()
	wordtags = multiset()
	sigs = multiset()
	sigtag = multiset()
	words = multiset(word for sent in tagged_sents for word, tag in sent)
	lexicon = {word for word, freq in words.items() if freq > unknownthreshold}
	wordsig = {}
	for sent in tagged_sents:
		for n, (word, tag) in enumerate(sent):
			wordsfortag[tag].add(word)
			tags[tag] += 1
			wordtags[word, tag] += 1
			sig = unknownword(word, n, lexicon)
			wordsig[word] = sig  # NB: sig may also depend on n and lexicon
			sigtag[sig, tag] += 1
	if openclassthreshold:
		openclasstags = {tag: len({w.lower() for w in ws})
				for tag, ws in wordsfortag.items()
				if len({w.lower() for w in ws}) >= openclassthreshold}
		closedclasstags = set(tags) - set(openclasstags)
		closedclasswords = {word for tag in closedclasstags
				for word in wordsfortag[tag]}
		openclasswords = lexicon - closedclasswords
		# add rare closed-class words back to lexicon
		lexicon.update(closedclasswords)
	else:
		openclasstags = {}
		openclasswords = {}
	for sent in tagged_sents:
		for n, (word, _) in enumerate(sent):
			if word not in lexicon:
				sig = unknownword(word, n, lexicon)
				sigs[sig] += 1
	msg = "known words: %d, signature types seen: %d\n" % (
			len(lexicon), len(sigs))
	msg += "open class tags: {%s}" % ", ".join(
			"%s:%d" % a for a in openclasstags.items())
	return (sigs, words, lexicon, wordsfortag, openclasstags,
			openclasswords, tags, wordtags,
			wordsig, sigtag), msg


def replaceraretrainwords(tagged_sents, unknownword, lexicon):
	""" Given a training set, replace all terminals not part of the lexicon
	with a signature of features as returned by unknownword(),
	before a grammar is read of from the training set. """
	return [[word if word in lexicon else unknownword(word, n, lexicon)
			for n, (word, _) in enumerate(sent)]
			for sent in tagged_sents]


def replaceraretestwords(sent, unknownword, lexicon, sigs):
	""" Given a test sentence, replace words not part of the lexicon
	with a signature of features as returned by unknownword(),
	if that signature is part of the grammar. """
	for n, word in enumerate(sent):
		if word in lexicon:
			yield word
		else:
			sig = unknownword(word, n, lexicon)
			if sig in sigs:
				yield sig
			else:
				yield UNK


def simplesmoothlexicon(lexmodel, epsilon=Fraction(1, 100)):
	""" introduce lexical productions for unobserved
	combinations of known open class words and tags, as well as for unobserved
	signatures which are mapped to '_UNK'.

	:param epsilon: 'frequency' of productions for unseen tag, word pair. """
	(lexicon, wordsfortag, openclasstags,
			openclasswords, tags, wordtags) = lexmodel
	newrules = []
	# rare words as signature AND as word:
	for word, tag in wordtags:
		if word not in lexicon:  # and word in openclasswords:
			# needs to be normalized later
			newrules.append((((tag, 'Epsilon'), (word, )),
					Fraction(wordtags[word, tag], tags[tag])))
			#print(>> sys.stderr, newrules[-1])
	for tag in openclasstags:  # open class tag-word pairs
		epsilon1 = epsilon / tags[tag]
		for word in openclasswords - wordsfortag[tag] - {UNK}:
			newrules.append((((tag, 'Epsilon'), (word, )), epsilon1))
	for tag in tags:  # catch all unknown signature
		epsilon1 = epsilon / tags[tag]
		newrules.append((((tag, 'Epsilon'), (UNK, )), epsilon1))
	return newrules


def getlexmodel(sigs, words, _lexicon, wordsfortag, openclasstags,
			openclasswords, tags, wordtags, wordsig, sigtag,
			openclassoffset=1, kappa=1):
	""" Compute a smoothed lexical model.

	:returns: a dictionary giving P(word_or_sig | tag).
	:param openclassoffset: for words that only appear with open class tags,
		add unseen combinations of open class (tag, word) with this count.
	:param kappa: FIXME; cf. Klein & Manning (2003). """
	for tag in openclasstags:
		for word in openclasswords - wordsfortag[tag]:
			wordtags[word, tag] += openclassoffset
			words[word] += openclassoffset
			tags[tag] += openclassoffset
		# unseen signatures
		sigs[UNK] += 1
		sigtag[UNK, tag] += 1
	# Compute P(tag|sig)
	tagtotal = sum(tags.values())
	wordstotal = sum(words.values())
	sigstotal = sum(sigs.values())
	P_tag = {}
	for tag in tags:
		P_tag[tag] = Fraction(tags[tag], tagtotal)
	P_word = defaultdict(int)
	for word in words:
		P_word[word] = Fraction(words[word], wordstotal)
	P_tagsig = defaultdict(Fraction)  # ??
	for sig in sigs:
		P_tagsig[tag, sig] = Fraction(P_tag[tag],
				Fraction(sigs[sig], sigstotal))
		#print(>> sys.stderr, "P(%s | %s) = %s " % ()
		#		tag, sig, P_tagsig[tag, sig])
	# Klein & Manning (2003) Accurate unlexicalized parsing
	# P(tag|word) = [count(tag, word) + kappa * P(tag|sig)]
	#		/ [count(word) + kappa]
	P_tagword = defaultdict(int)
	for word, tag in wordtags:
		P_tagword[tag, word] = Fraction(wordtags[word, tag]
				+ kappa * P_tagsig[tag, wordsig[word]],
				words[word] + kappa)
		#print(>> sys.stderr, "P(%s | %s) = %s " % ()
		#		tag, word, P_tagword[tag, word])
	# invert with Bayes theorem to get P(word|tag)
	P_wordtag = defaultdict(int)
	for tag, word in P_tagword:
		#wordorsig = word if word in lexicon else wordsig[word]
		wordorsig = word
		P_wordtag[wordorsig, tag] += Fraction((P_tagword[tag, word]
				* P_word[word]), P_tag[tag])
		#print(>> sys.stderr, "P(%s | %s) = %s " % ()
		#		word, tag, P_wordtag[wordorsig, tag])
	msg = "(word, tag) pairs in model: %d" % len(P_tagword)
	return P_wordtag, msg


def smoothlexicon(grammar, P_wordtag):
	""" Replace lexical probabilities using given unknown word model.
	Ignores lexical productions of known subtrees (tag contains '@')
	introduced by DOP, i.e., we only modify lexical depth 1 subtrees. """
	newrules = []
	for (rule, yf), w in grammar:
		if rule[1] == 'Epsilon' and '@' not in rule[0]:
			wordorsig = yf[0]
			tag = rule[0]
			newrule = (((tag, 'Epsilon'), (wordorsig, )),
					P_wordtag[wordorsig, tag])
			newrules.append(newrule)
		else:
			newrules.append(((rule, yf), w))
	return newrules


# === functions for unknown word signatures ============
# resolve name to function assigning unknown word signatures
def getunknownwordfun(model):
	""" Resolve name of unknown word model into function for that model. """
	return {"4": unknownword4,
		"6": unknownword6,
		"base": unknownwordbase,
		"ftb": unknownwordftb,
		}[model]

hasdigit = re.compile(r"\d", re.UNICODE)
hasnondigit = re.compile(r"\D", re.UNICODE)
#NB: includes '-', hyphen, non-breaking hyphen
# does NOT include: figure-dash, em-dash, en-dash (these are punctuation,
# not word-combining) u2012-u2015; nb: these are hex values.
hasdash = re.compile(u"[-\u2010\u2011]", re.UNICODE)
# FIXME: exclude accented characters for model 6?
haslower = re.compile(u'[a-z\xe7\xe9\xe0\xec\xf9\xe2\xea\xee\xf4\xfb\xeb'
		u'\xef\xfc\xff\u0153\xe6]', re.UNICODE)
hasupper = re.compile(u'[A-Z\xc7\xc9\xc0\xcc\xd9\xc2\xca\xce\xd4\xdb\xcb'
		u'\xcf\xdc\u0178\u0152\xc6]', re.UNICODE)
hasletter = re.compile(u'[A-Za-z\xe7\xe9\xe0\xec\xf9\xe2\xea\xee\xf4\xfb'
		u'\xeb\xef\xfc\xff\u0153\xe6\xc7\xc9\xc0\xcc\xd9\xc2\xca\xce\xd4'
		u'\xdb\xcb\xcf\xdc\u0178\u0152\xc6]', re.UNICODE)
# Cf. http://en.wikipedia.org/wiki/French_alphabet
LOWER = (u'abcdefghijklmnopqrstuvwxyz\xe7\xe9\xe0\xec\xf9\xe2\xea\xee\xf4\xfb'
		u'\xeb\xef\xfc\xff\u0153\xe6')
UPPER = (u'ABCDEFGHIJKLMNOPQRSTUVWXYZ\xc7\xc9\xc0\xcc\xd9\xc2\xca\xce\xd4\xdb'
		u'\xcb\xcf\xdc\u0178\u0152\xc6')
LOWERUPPER = LOWER + UPPER


def unknownword6(word, loc, lexicon):
	""" Model 6 of the Stanford parser (for WSJ treebank). """
	wlen = len(word)
	numcaps = 0
	sig = UNK
	numcaps = len(hasupper.findall(word))
	lowered = word.lower()
	if numcaps > 1:
		sig += "-CAPS"
	elif numcaps > 0:
		if loc == 0:
			sig += "-INITC"
			if lowered in lexicon:
				sig += "-KNOWNLC"
		else:
			sig += "-CAP"
	elif haslower.search(word):
		sig += "-LC"
	if hasdigit.search(word):
		sig += "-NUM"
	if hasdash.search(word):
		sig += "-DASH"
	if lowered.endswith("s") and wlen >= 3:
		if lowered[-2] not in "siu":
			sig += "-s"
	elif wlen >= 5 and not hasdash.search(word) and not (
			hasdigit.search(word) and numcaps > 0):
		suffixes = ("ed", "ing", "ion", "er", "est", "ly", "ity", "y", "al")
		for a in suffixes:
			if lowered.endswith(a):
				sig += "-%s" % a
	return sig


def unknownword4(word, loc, _):
	""" Model 4 of the Stanford parser. Relatively language agnostic. """
	sig = UNK

	# letters
	if word[0] in UPPER:
		if not haslower:
			sig += "-AC"
		elif loc == 0:
			sig += "-SC"
		else:
			sig += "-C"
	elif haslower.search(word):
		sig += "-L"
	elif hasletter.search(word):
		sig += "-U"
	else:
		sig += "-S"  # no letter

	# digits
	if hasdigit.search(word):
		if hasnondigit.search(word):
			sig += "-n"
		else:
			sig += "-N"

	# punctuation
	if "-" in word:
		sig += "-H"
	if "." in word:
		sig += "-P"
	if "," in word:
		sig += "-C"
	if len(word) > 3:
		if word[-1] in LOWERUPPER:
			sig += "-%s" % word[-1].lower()
	return sig


def unknownwordbase(word, _loc, _lexicon):
	""" BaseUnknownWordModel of the Stanford parser.
	Relatively language agnostic. """
	sig = UNK

	# letters
	if word[0] in UPPER:
		sig += "-C"
	else:
		sig += "-c"

	# digits
	if hasdigit.search(word):
		if hasnondigit.search(word):
			sig += "-n"
		else:
			sig += "-N"

	# punctuation
	if "-" in word:
		sig += "-H"
	if word == ".":
		sig += "-P"
	if word == ",":
		sig += "-C"
	if len(word) > 3:
		if word[-1] in LOWERUPPER:
			sig += "-%s" % word[-2].lower()
	return sig

nounsuffix = re.compile(u"(ier|ière|ité|ion|ison|isme|ysme|iste|esse|eur|euse"
		u"|ence|eau|erie|ng|ette|age|ade|ance|ude|ogue|aphe|ate|duc|anthe"
		u"|archie|coque|érèse|ergie|ogie|lithe|mètre|métrie|odie|pathie|phie"
		u"|phone|phore|onyme|thèque|scope|some|pole|ôme|chromie|pie)s?$")
adjsuffix = re.compile(u"(iste|ième|uple|issime|aire|esque|atoire|ale|al|able"
		u"|ible|atif|ique|if|ive|eux|aise|ent|ois|oise|ante|el|elle|ente|oire"
		u"|ain|aine)s?$")
possibleplural = re.compile(u"(s|ux)$")
verbsuffix = re.compile(u"(ir|er|re|ez|ont|ent|ant|ais|ait|ra|era|eras"
		u"|é|és|ées|isse|it)$")
advsuffix = re.compile("(iment|ement|emment|amment)$")
haspunc = re.compile(u"([\u0021-\u002F\u003A-\u0040\u005B\u005C\u005D"
		u"\u005E-\u0060\u007B-\u007E\u00A1-\u00BF\u2010-\u2027\u2030-\u205E"
		u"\u20A0-\u20B5])+")
ispunc = re.compile(u"([\u0021-\u002F\u003A-\u0040\u005B\u005C\u005D"
		u"\u005E-\u0060\u007B-\u007E\u00A1-\u00BF\u2010-\u2027\u2030-\u205E"
		u"\u20A0-\u20B5])+$")


def unknownwordftb(word, loc, _):
	""" Model 2 for French of the Stanford parser. """
	sig = UNK

	if advsuffix.search(word):
		sig += "-ADV"
	elif verbsuffix.search(word):
		sig += "-VB"
	elif nounsuffix.search(word):
		sig += "-NN"

	if adjsuffix.search(word):
		sig += "-ADV"
	if hasdigit.search(word):
		sig += "-NUM"
	if possibleplural.search(word):
		sig += "-PL"

	if ispunc.search(word):
		sig += "-ISPUNC"
	elif haspunc.search(word):
		sig += "-HASPUNC"

	if loc > 0 and len(word) > 0 and word[0] in UPPER:
		sig += "-UP"

	return sig


def main():
	""" Not implemented. """

if __name__ == '__main__':
	main()
